"""The formatting module includes functions to apply IRC formatting to text.

*Availability: 4.5+*
"""
# Copyright 2014, Elsie Powell, embolalia.com
# Copyright 2019, dgw, technobabbl.es
# Licensed under the Eiffel Forum License 2.
from __future__ import annotations

from enum import Enum
import re
import string


__all__ = [
    # control chars
    'CONTROL_NORMAL',
    'CONTROL_COLOR',
    'CONTROL_HEX_COLOR',
    'CONTROL_BOLD',
    'CONTROL_ITALIC',
    'CONTROL_UNDERLINE',
    'CONTROL_STRIKETHROUGH',
    'CONTROL_MONOSPACE',
    'CONTROL_REVERSE',
    # convenience lists
    'CONTROL_FORMATTING',
    'CONTROL_NON_PRINTING',
    # utility functions
    'color',
    'hex_color',
    'bold',
    'italic',
    'underline',
    'strikethrough',
    'monospace',
    'reverse',
    'plain',
    # utility enum
    'colors',
]

# Color names are as specified at http://www.mirc.com/colors.html

CONTROL_NORMAL = '\x0f'
"""The control code to reset formatting."""
CONTROL_COLOR = '\x03'
"""The control code to start or end color formatting."""
CONTROL_HEX_COLOR = '\x04'
"""The control code to start or end hexadecimal color formatting."""
CONTROL_BOLD = '\x02'
"""The control code to start or end bold formatting."""
CONTROL_ITALIC = '\x1d'
"""The control code to start or end italic formatting."""
CONTROL_UNDERLINE = '\x1f'
"""The control code to start or end underlining."""
CONTROL_STRIKETHROUGH = '\x1e'
"""The control code to start or end strikethrough formatting."""
CONTROL_MONOSPACE = '\x11'
"""The control code to start or end monospace formatting."""
CONTROL_REVERSE = '\x16'
"""The control code to start or end reverse-color formatting."""

CONTROL_FORMATTING = [
    CONTROL_NORMAL,
    CONTROL_COLOR,
    CONTROL_HEX_COLOR,
    CONTROL_BOLD,
    CONTROL_ITALIC,
    CONTROL_UNDERLINE,
    CONTROL_STRIKETHROUGH,
    CONTROL_MONOSPACE,
    CONTROL_REVERSE,
]
"""A list of all control characters expected to appear as formatting."""

CONTROL_NON_PRINTING = [
    '\x00',
    '\x01',
    '\x02',  # CONTROL_BOLD
    '\x03',  # CONTROL_COLOR
    '\x04',  # CONTROL_HEX_COLOR
    '\x05',
    '\x06',
    '\x07',
    '\x08',
    '\x09',
    '\x0a',
    '\x0b',
    '\x0c',
    '\x0d',
    '\x0e',
    '\x0f',  # CONTROL_NORMAL
    '\x10',
    '\x11',  # CONTROL_MONOSPACE
    '\x12',
    '\x13',
    '\x14',
    '\x15',
    '\x16',  # CONTROL_REVERSE
    '\x17',
    '\x18',
    '\x19',
    '\x1a',
    '\x1b',
    '\x1c',
    '\x1d',  # CONTROL_ITALIC
    '\x1e',  # CONTROL_STRIKETHROUGH
    '\x1f',  # CONTROL_UNDERLINE
    '\x7f',
]

# Regex to detect Control Pattern
COLOR_PATTERN = re.escape(CONTROL_COLOR) + r'((\d{1,2},\d{2})|\d{2})?'
HEX_COLOR_PATTERN = '%s(%s)?' % (
    re.escape(CONTROL_HEX_COLOR),
    '|'.join([
        '(' + ','.join([r'[a-fA-F0-9]{6}', r'[a-fA-F0-9]{6}']) + ')',
        r'[a-fA-F0-9]{6}'
    ])
)

PLAIN_PATTERN = '|'.join([
    '(' + COLOR_PATTERN + ')',
    '(' + HEX_COLOR_PATTERN + ')',
])
PLAIN_REGEX = re.compile(PLAIN_PATTERN)


class colors(str, Enum):
    """Mapping of color names to mIRC code values."""
    # Mostly aligned with https://modern.ircdocs.horse/formatting.html#colors
    # which are likely based on mIRC's color names (https://www.mirc.com/colors.html)
    WHITE = '00'
    BLACK = '01'
    BLUE = '02'
    GREEN = '03'
    LIGHT_RED = '04'
    BROWN = '05'
    PURPLE = '06'
    ORANGE = '07'
    YELLOW = '08'
    LIGHT_GREEN = '09'
    TEAL = '10'  # TODO: should be called 'CYAN'
    LIGHT_CYAN = '11'
    LIGHT_BLUE = '12'
    PINK = '13'
    GREY = '14'
    LIGHT_GREY = '15'

    # Create aliases.
    NAVY = BLUE
    RED = LIGHT_RED
    MAROON = BROWN
    OLIVE = ORANGE  # TODO: held over from antiquity; does anyone actually think this?
    LIME = LIGHT_GREEN
    CYAN = LIGHT_CYAN  # TODO: should be what is called 'TEAL' above
    ROYAL = LIGHT_BLUE
    LIGHT_PURPLE = PINK
    FUCHSIA = PINK
    GRAY = GREY
    LIGHT_GRAY = LIGHT_GREY
    SILVER = LIGHT_GREY


def _get_color(color):
    if color is None:
        return None

    # You can pass an int or string of the code
    try:
        color = int(color)
    except ValueError:
        pass
    if isinstance(color, int):
        if color > 99:
            raise ValueError('Can not specify a color above 99.')
        return str(color).rjust(2, '0')

    # You can also pass the name of the color
    color_name = color.upper()
    color_dict = colors.__dict__
    try:
        return color_dict[color_name]
    except KeyError:
        raise ValueError('Unknown color name {}'.format(color))


def color(text, fg=None, bg=None):
    """Return the text, with the given colors applied in IRC formatting.

    :param str text: the text to format
    :param mixed fg: the foreground color
    :param mixed bg: the background color
    :raises TypeError: if ``text`` is not a string
    :raises ValueError: if ``fg`` or ``bg`` is an unrecognized color value
    :rtype: str

    The color can be a string of the color name, or an integer in the range
    0-99. The known color names can be found in the :class:`colors` class of
    this module.
    """
    if fg is None and bg is None:
        return text

    fg = _get_color(fg)
    bg = _get_color(bg)

    if not bg:
        text = ''.join([CONTROL_COLOR, fg, text, CONTROL_COLOR])
    else:
        text = ''.join([CONTROL_COLOR, fg, ',', bg, text, CONTROL_COLOR])
    return text


def _get_hex_color(color):
    if color is None:
        return None

    try:
        color = color.upper()
        if not all(c in string.hexdigits for c in color):
            raise AttributeError
    except AttributeError:
        raise ValueError('Hexadecimal color value must be passed as string.')

    if len(color) == 3:
        return ''.join([c * 2 for c in color])
    elif len(color) == 6:
        return color
    else:  # invalid length
        raise ValueError('Hexadecimal color value must have either 3 or 6 digits.')


def hex_color(text, fg=None, bg=None):
    """Return the text, with the given colors applied in IRC formatting.

    :param str text: the text to format
    :param str fg: the foreground color
    :param str bg: the background color
    :raises TypeError: if ``text`` is not a string
    :raises ValueError: if ``fg`` or ``bg`` is an unrecognized color value
    :rtype: str

    The color can be provided with a string of either 3 or 6 hexadecimal digits.
    As in CSS, 3-digit colors will be interpreted as if they were 6-digit colors
    with each digit repeated (e.g. color ``c90`` is identical to ``cc9900``). Do
    not include the leading ``#`` symbol.

    .. note::
        This is a relatively new IRC formatting convention. Use it only when you
        can afford to have its meaning lost, as not many clients support it yet.
    """
    if not fg and not bg:
        return text

    fg = _get_hex_color(fg)
    bg = _get_hex_color(bg)

    if not bg:
        text = ''.join([CONTROL_HEX_COLOR, fg, text, CONTROL_HEX_COLOR])
    else:
        text = ''.join([CONTROL_HEX_COLOR, fg, ',', bg, text, CONTROL_HEX_COLOR])
    return text


def bold(text):
    """Return the text, with bold IRC formatting.

    :param str text: the text to format
    :raises TypeError: if ``text`` is not a string
    :rtype: str
    """
    return ''.join([CONTROL_BOLD, text, CONTROL_BOLD])


def italic(text):
    """Return the text, with italic IRC formatting.

    :param str text: the text to format
    :raises TypeError: if ``text`` is not a string
    :rtype: str
    """
    return ''.join([CONTROL_ITALIC, text, CONTROL_ITALIC])


def underline(text):
    """Return the text, with underline IRC formatting.

    :param str text: the text to format
    :raises TypeError: if ``text`` is not a string
    :rtype: str
    """
    return ''.join([CONTROL_UNDERLINE, text, CONTROL_UNDERLINE])


def strikethrough(text):
    """Return the text, with strikethrough IRC formatting.

    :param str text: the text to format
    :raises TypeError: if ``text`` is not a string
    :rtype: str

    .. note::
        This is a relatively new IRC formatting convention. Use it only when you
        can afford to have its meaning lost, as not many clients support it yet.
    """
    return ''.join([CONTROL_STRIKETHROUGH, text, CONTROL_STRIKETHROUGH])


def monospace(text):
    """Return the text, with monospace IRC formatting.

    :param str text: the text to format
    :raises TypeError: if ``text`` is not a string
    :rtype: str

    .. note::
        This is a relatively new IRC formatting convention. Use it only when you
        can afford to have its meaning lost, as not many clients support it yet.
    """
    return ''.join([CONTROL_MONOSPACE, text, CONTROL_MONOSPACE])


def reverse(text):
    """Return the text, with reverse-color IRC formatting.

    :param str text: the text to format
    :raises TypeError: if ``text`` is not a string
    :rtype: str

    .. note::
        This code isn't super well supported, and its behavior even in clients
        that understand it (e.g. mIRC) can be unpredictable. Use it carefully.
    """
    return ''.join([CONTROL_REVERSE, text, CONTROL_REVERSE])


def plain(text):
    """Return the text without any IRC formatting.

    :param str text: text with potential IRC formatting control code(s)
    :raises TypeError: if ``text`` is not a string
    :rtype: str
    """
    if '\x03' in text or '\x04' in text:
        text = PLAIN_REGEX.sub('', text)
    return ''.join(c for c in text if ord(c) >= 0x20 and c != '\x7F')



"""Identifier tools to represent IRC names (nick or channel).

Nick and channel are defined by their names, which are "identifiers": their
names are used to differentiate users from each others, channels from each
others. To ensure that two channels or two users are the same, their
identifiers must be processed to be compared properly. This process depends on
which RFC and how that RFC is implemented by the server: IRC being an old
protocol, different RFCs have different versions of that process:

* :rfc:`1459#section-2.2`: ASCII characters, and ``[]\\`` are mapped to ``{}|``
* :rfc:`2812#section-2.2`: same as in the previous RFC, adding ``~`` mapped to
  ``^``

Then when ISUPPORT was added, the `CASEMAPPING parameter`__ was defined so the
server can say which process to apply:

* ``ascii``: only ``[A-Z]`` must be mapped to ``[a-z]`` (implemented by
  :func:`ascii_lower`)
* ``rfc1459``: follows :rfc:`2812`, because of how it was implemented in most
  servers (implemented by :func:`rfc1459_lower`)
* A strict version of :rfc:`1459` also exists but it is not recommended
  (implemented by :func:`rfc1459_strict_lower`)

As a result, the :class:`Identifier` class requires a casemapping function,
which should be provided by the :class:`bot<sopel.bot.Sopel>`.

.. seealso::

    The bot's :class:`make_identifier<sopel.bot.Sopel.make_identifier>` method
    should be used to instantiate an :class:`Identifier` to honor the
    ``CASEMAPPING`` parameter.

.. __: https://modern.ircdocs.horse/index.html#casemapping-parameter
"""
from __future__ import annotations

import string
from typing import Callable

Casemapping = Callable[[str], str]
"""Type definition of a casemapping callable."""

ASCII_TABLE = str.maketrans(string.ascii_uppercase, string.ascii_lowercase)
RFC1459_TABLE = str.maketrans(
    string.ascii_uppercase + '[]\\~',
    string.ascii_lowercase + '{}|^',
)
RFC1459_STRICT_TABLE = str.maketrans(
    string.ascii_uppercase + '[]\\',
    string.ascii_lowercase + '{}|',
)
DEFAULT_CHANTYPES = ('#', '&', '+', '!')


def ascii_lower(text: str) -> str:
    """Lower ``text`` according to the ``ascii`` value of ``CASEMAPPING``.

    In that version, only ``[A-Z]`` are to be mapped to their lowercase
    equivalent (``[a-z]``). Non-ASCII characters are kept unmodified.
    """
    return text.translate(ASCII_TABLE)


def rfc1459_lower(text: str) -> str:
    """Lower ``text`` according to :rfc:`2812`.

    Similar to :func:`rfc1459_strict_lower`, but also maps ``~`` to ``^``, as
    per :rfc:`2812#section-2.2`:

        Because of IRC's Scandinavian origin, the characters ``{}|^`` are
        considered to be the lower case equivalents of the characters
        ``[]\\~``, respectively.

    .. note::

        This is an implementation of the `CASEMAPPING parameter`__ for the
        value ``rfc1459``, which doesn't use :rfc:`1459` but its updated version
        :rfc:`2812`.

    .. __: https://modern.ircdocs.horse/index.html#casemapping-parameter
    """
    return text.translate(RFC1459_TABLE)


def rfc1459_strict_lower(text: str) -> str:
    """Lower ``text`` according to :rfc:`1459` (strict version).

    As per :rfc:`1459#section-2.2`:

        Because of IRC's scandanavian origin, the characters ``{}|`` are
        considered to be the lower case equivalents of the characters ``[]\\``.

    """
    return text.translate(RFC1459_STRICT_TABLE)


class Identifier(str):
    """A ``str`` subclass which acts appropriately for IRC identifiers.

    :param identifier: IRC identifier
    :param casemapping: a casemapping function (optional keyword argument)

    When used as normal ``str`` objects, case will be preserved.
    However, when comparing two Identifier objects, or comparing an Identifier
    object with a ``str`` object, the comparison will be case insensitive.

    This case insensitivity uses the provided ``casemapping`` function,
    following the rules for the `CASEMAPPING parameter`__ from ISUPPORT. By
    default, it uses :func:`rfc1459_lower`, following :rfc:`2812#section-2.2`.

    .. note::

        To instantiate an ``Identifier`` with the appropriate ``casemapping``
        function, it is best to rely on
        :meth:`bot.make_identifier<sopel.irc.AbstractBot.make_identifier>`.

    .. versionchanged:: 8.0

        The ``casemapping`` and ``chantypes`` parameters have been added.

    .. __: https://modern.ircdocs.horse/index.html#casemapping-parameter
    """
    def __new__(
        cls,
        identifier: str,
        *,
        casemapping: Casemapping = rfc1459_lower,
        chantypes: tuple = DEFAULT_CHANTYPES,
    ) -> 'Identifier':
        return str.__new__(cls, identifier)

    def __init__(
        self,
        identifier: str,
        *,
        casemapping: Casemapping = rfc1459_lower,
        chantypes: tuple = DEFAULT_CHANTYPES,
    ) -> None:
        super().__init__()
        self.casemapping: Casemapping = casemapping
        """Casemapping function to lower the identifier."""
        self.chantypes = chantypes
        """Tuple of prefixes used for channels."""
        self._lowered = self.casemapping(identifier)

    def lower(self) -> str:
        """Get the IRC-compliant lowercase version of this identifier.

        :return: IRC-compliant lowercase version used for case-insensitive
                 comparisons

        The behavior of this method depends on the identifier's casemapping
        function, which should be selected based on the ``CASEMAPPING``
        parameter from ``ISUPPORT``.

        .. versionchanged:: 8.0

            Now uses the :attr:`casemapping` function to lower the identifier.

        """
        return self.casemapping(self)

    @staticmethod
    def _lower(identifier: str) -> str:
        """Convert an identifier to lowercase per :rfc:`2812`.

        :param identifier: the identifier (nickname or channel) to convert
        :return: RFC 2812-compliant lowercase version of ``identifier``

        :meta public:

        .. versionchanged:: 8.0

            Previously, this would lower all non-ASCII characters. It now uses
            a strict implementation of the ``CASEMAPPING`` parameter. This is
            now equivalent to call :func:`rfc1459_lower`.

            If the ``identifier`` is an instance of :class:`Identifier`, this
            will call that identifier's :meth:`lower` method instead.

        """
        if isinstance(identifier, Identifier):
            return identifier.lower()
        return rfc1459_lower(identifier)

    @staticmethod
    def _lower_swapped(identifier: str) -> str:
        """Backward-compatible version of :meth:`_lower`.

        :param identifier: the identifier (nickname or channel) to convert
        :return: RFC 2812-non-compliant lowercase version of ``identifier``

        This is what the old :meth:`_lower` function did before Sopel 7.0. It
        maps ``{}``, ``[]``, ``|``, ``\\``, ``^``, and ``~`` incorrectly.

        You shouldn't use this unless you need to migrate stored values from
        the previous, incorrect "lowercase" representation to the correct one.

        :meta public:

        .. versionadded: 7.0

            This method was added to ensure migration of improperly lowercased
            data: it reverts the data back to the previous lowercase rules.

        """
        # The tilde replacement isn't needed for identifiers, but is for
        # channels, which may be useful at some point in the future.
        # Always convert to str, to prevent using custom casemapping
        low = str(identifier).lower().replace('{', '[').replace('}', ']')
        low = low.replace('|', '\\').replace('^', '~')
        return low

    def __repr__(self):
        return "%s(%r)" % (
            self.__class__.__name__,
            self.__str__()
        )

    def __hash__(self):
        return self._lowered.__hash__()

    def __lt__(self, other):
        if isinstance(other, str):
            other = self.casemapping(other)
        return str.__lt__(self._lowered, other)

    def __le__(self, other):
        if isinstance(other, str):
            other = self.casemapping(other)
        return str.__le__(self._lowered, other)

    def __gt__(self, other):
        if isinstance(other, str):
            other = self.casemapping(other)
        return str.__gt__(self._lowered, other)

    def __ge__(self, other):
        if isinstance(other, str):
            other = self.casemapping(other)
        return str.__ge__(self._lowered, other)

    def __eq__(self, other):
        if isinstance(other, str):
            other = self.casemapping(other)
        return str.__eq__(self._lowered, other)

    def __ne__(self, other):
        return not (self == other)

    def is_nick(self) -> bool:
        """Check if the Identifier is a nickname (i.e. not a channel)

        :return: ``True`` if this :py:class:`Identifier` is a nickname;
                 ``False`` if it appears to be a channel

        ::

            >>> from sopel import tools
            >>> ident = tools.Identifier('Sopel')
            >>> ident.is_nick()
            True
            >>> ident = tools.Identifier('#sopel')
            >>> ident.is_nick()
            False

        To detect channels, :attr:`chantypes` is used::

            >>> from sopel import tools
            >>> ident = tools.Identifier('!sopel', chantypes=('#', '&'))
            >>> ident.is_nick()
            True
            >>> ident.chantypes = ('#', '&', '!')
            >>> ident.is_nick()
            False

        """
        return bool(self) and not self.startswith(self.chantypes)


IdentifierFactory = Callable[[str], Identifier]
"""Type definition of an identifier factory."""



"""
tell.py - Sopel Tell and Ask Plugin
Copyright 2008, Sean B. Palmer, inamidst.com
Copyright 2019, dgw, technobabbl.es
Licensed under the Eiffel Forum License 2.

https://sopel.chat
"""
from __future__ import annotations

from collections import defaultdict
import io  # don't use `codecs` for loading the DB; it will split lines on some IRC formatting
import logging
import os
import threading
import time
import unicodedata

from sopel import formatting, plugin
from sopel.config import types
from sopel.tools.time import format_time, get_timezone


LOGGER = logging.getLogger(__name__)


class TellSection(types.StaticSection):
    use_private_reminder = types.BooleanAttribute(
        'use_private_reminder', default=False)
    """When set to ``true``, Sopel will send reminder as private message."""
    maximum_public = types.ValidatedAttribute(
        'maximum_public', parse=int, default=4)
    """How many Sopel can send in public before using private message."""


def configure(config):
    """
    | name | example | purpose |
    | ---- | ------- | ------- |
    | use_private_reminder | false | Send reminders as private message |
    | maximum_public | 4 | Send up to this amount of reminders in public |
    """
    config.define_section('tell', TellSection)
    config.tell.configure_setting(
        'use_private_reminder',
        'Should Sopel send tell/ask reminders as private message only?')
    if not config.tell.use_private_reminder:
        config.tell.configure_setting(
            'maximum_public',
            'How many tell/ask reminders Sopel will send as public message '
            'before sending them as private messages?')


def load_reminders(filename):
    """Load tell/ask reminders from a ``filename``.

    :param str filename: path to the tell/ask reminders file
    :return: a dict with the tell/ask reminders
    :rtype: dict
    """
    result = defaultdict(list)
    with io.open(filename, 'r', encoding='utf-8') as fd:
        for line in fd:
            line = line.strip()
            if line:
                try:
                    tellee, teller, verb, timenow, msg = line.split('\t', 4)
                except ValueError:
                    continue  # TODO: Add warning log about malformed reminder
                result[tellee].append((teller, verb, timenow, msg))

    return result


def dump_reminders(filename, data):
    """Dump tell/ask reminders (``data``) into a ``filename``.

    :param str filename: path to the tell/ask reminders file
    :param dict data: tell/ask reminders ``dict``
    """
    with io.open(filename, 'w', encoding='utf-8') as fd:
        for tellee, reminders in data.items():
            for reminder in reminders:
                line = '\t'.join((tellee,) + tuple(reminder))
                fd.write(line + '\n')
    return True


def setup(bot):
    bot.config.define_section('tell', TellSection)
    fn = bot.config.basename + '.tell.db'
    bot.tell_filename = os.path.join(bot.config.core.homedir, fn)

    # Pre-7.0 migration logic. Remove in 8.0 or 9.0.
    old = bot.nick + '-' + bot.config.core.host + '.tell.db'
    old = os.path.join(bot.config.core.homedir, old)
    if os.path.isfile(old):
        LOGGER.info("Attempting to migrate old 'tell' database {}..."
                    .format(old))
        try:
            os.rename(old, bot.tell_filename)
        except OSError:
            LOGGER.error("Migration failed!")
            LOGGER.error("Old filename: {}".format(old))
            LOGGER.error("New filename: {}".format(bot.tell_filename))
            LOGGER.error(
                "See https://sopel.chat/usage/installing/upgrading-to-sopel-7/#reminder-db-migration")
        else:
            LOGGER.info("Migration finished!")
    # End migration logic

    if not os.path.exists(bot.tell_filename):
        with io.open(bot.tell_filename, 'w', encoding='utf-8') as fd:
            # if we can't open/write into the file, the tell plugin can't work
            fd.write('')

    if 'tell_lock' not in bot.memory:
        bot.memory['tell_lock'] = threading.Lock()

    if 'reminders' not in bot.memory:
        with bot.memory['tell_lock']:
            bot.memory['reminders'] = load_reminders(bot.tell_filename)


def shutdown(bot):
    for key in ['tell_lock', 'reminders']:
        try:
            del bot.memory[key]
        except KeyError:
            pass


def _format_safe_lstrip(text):
    """``str.lstrip()`` but without eating IRC formatting.

    :param str text: text to clean
    :rtype: str
    :raises TypeError: if the passed ``text`` is not a string

    Stolen and tweaked from the ``choose`` plugin's ``_format_safe()``
    function by the person who wrote it.
    """
    if not isinstance(text, str):
        raise TypeError("A string is required.")
    elif not text:
        # unnecessary optimization
        return ''

    start = 0

    # strip left
    pos = 0
    while pos < len(text):
        is_whitespace = unicodedata.category(text[pos]) == 'Zs'
        is_non_printing = (
            text[pos] in formatting.CONTROL_NON_PRINTING and
            text[pos] not in formatting.CONTROL_FORMATTING
        )
        if not is_whitespace and not is_non_printing:
            start = pos
            break
        pos += 1
    else:
        # skipped everything; string is all whitespace
        return ''

    return text[start:]


@plugin.command('tell', 'ask')
@plugin.nickname_command('tell', 'ask')
@plugin.example('$nickname, tell dgw he broke it again.', user_help=True)
@plugin.example('.ask ', 'ask whom?')
@plugin.example('.tell ', 'tell whom?')
@plugin.example('.ask @', 'ask whom?')
@plugin.example('.tell @', 'tell whom?')
@plugin.example('.ask Exirel ', 'ask Exirel what?')
@plugin.example('.tell Exirel ', 'tell Exirel what?')
@plugin.example('.ask @Exirel ', 'ask Exirel what?')
@plugin.example('.tell @Exirel ', 'tell Exirel what?')
def f_remind(bot, trigger):
    """Give someone a message the next time they're seen"""
    teller = trigger.nick
    verb = trigger.group(1)

    if not trigger.group(3):
        tellee = ""
    else:
        tellee = trigger.group(3).rstrip('.,:;').lstrip('@')

    if not tellee:
        bot.reply("%s whom?" % verb)
        return

    # all we care about is having at least one non-whitespace
    # character after the name
    if not trigger.group(4):
        bot.reply("%s %s what?" % (verb, tellee))
        return

    msg = _format_safe_lstrip(trigger.group(2).split(' ', 1)[1])

    if not msg:
        bot.reply("%s %s what?" % (verb, tellee))
        return

    tellee = bot.make_identifier(tellee)

    if not os.path.exists(bot.tell_filename):
        return

    if len(tellee) > bot.isupport.get('NICKLEN', 30):
        bot.reply('That nickname is too long.')
        return

    if tellee == bot.nick:
        bot.reply("I'm here now; you can %s me whatever you want!" % verb)
        return

    if tellee not in (bot.make_identifier(teller), bot.nick, 'me'):
        tz = get_timezone(bot.db, bot.config, None, tellee)
        timenow = format_time(bot.db, bot.config, tz, tellee)
        with bot.memory['tell_lock']:
            if tellee not in bot.memory['reminders']:
                bot.memory['reminders'][tellee] = [(teller, verb, timenow, msg)]
            else:
                bot.memory['reminders'][tellee].append((teller, verb, timenow, msg))
            # save the reminders
            dump_reminders(bot.tell_filename, bot.memory['reminders'])

        response = "I'll pass that on when %s is around." % tellee
        bot.reply(response)
    elif bot.make_identifier(teller) == tellee:
        bot.reply('You can %s yourself that.' % verb)
    else:
        bot.reply("Hey, I'm not as stupid as Monty you know!")


def get_nick_reminders(reminders, nick):
    lines = []
    template = "%s: %s <%s> %s %s %s"
    today = time.strftime('%d %b', time.gmtime())

    for (teller, verb, datetime, msg) in reminders:
        if datetime.startswith(today):
            datetime = datetime[len(today) + 1:]
        lines.append(template % (nick, datetime, teller, verb, nick, msg))

    return lines


def nick_match_tellee(nick, tellee):
    """Tell if a ``nick`` matches a ``tellee``.

    :param str nick: Nick seen by the bot
    :param str tellee: Tellee name or pattern

    The check between ``nick`` and ``tellee`` is case-insensitive::

        >>> nick_match_tellee('Exirel', 'exirel')
        True
        >>> nick_match_tellee('exirel', 'EXIREL')
        True
        >>> nick_match_tellee('exirel', 'dgw')
        False

    If ``tellee`` ends with a wildcard token (``*`` or ``:``), then ``nick``
    matches if it starts with ``tellee`` (without the token)::

        >>> nick_match_tellee('Exirel', 'Exi*')
        True
        >>> nick_match_tellee('Exirel', 'exi:')
        True
        >>> nick_match_tellee('Exirel', 'Exi')
        False

    Note that this is still case-insensitive.
    """
    if tellee[-1] in ['*', ':']:  # these are wildcard token
        return nick.lower().startswith(tellee.lower().rstrip('*:'))
    return nick.lower() == tellee.lower()


@plugin.rule('(.*)')
@plugin.priority('low')
@plugin.unblockable
@plugin.output_prefix('[tell] ')
def message(bot, trigger):
    nick = trigger.nick

    if not os.path.exists(bot.tell_filename):
        # plugin can't work without its storage file
        return

    # get all matching reminders
    reminders = []
    tellees = list(reversed(sorted(
        tellee
        for tellee in bot.memory['reminders']
        if nick_match_tellee(nick, tellee)
    )))

    with bot.memory['tell_lock']:
        # pop reminders for nick
        reminders = list(
            reminder
            for tellee in tellees
            for reminder in get_nick_reminders(
                bot.memory['reminders'].pop(tellee, []), nick)
        )

    # check if there are reminders to send
    if not reminders:
        return  # nothing to do

    # then send reminders (as public and/or private messages)
    if bot.config.tell.use_private_reminder:
        # send reminders with private messages
        for line in reminders:
            bot.say(line, nick)
    else:
        # send up to 'maximum_public' reminders to the channel
        max_public = bot.config.tell.maximum_public
        for line in reminders[:max_public]:
            bot.say(line)

        # send other reminders directly to nick as private message
        if reminders[max_public:]:
            bot.reply('Further messages sent privately')
            for line in reminders[max_public:]:
                bot.say(line, nick)

    # save reminders left in memory
    with bot.memory['tell_lock']:
        dump_reminders(bot.tell_filename, bot.memory['reminders'])



"""Triggers are how Sopel tells callables about their runtime context.

A :class:`~.trigger.Trigger` is the main type of user input plugins will see.

Sopel uses :class:`~.trigger.PreTrigger`\\s internally while processing
incoming IRC messages. Plugin authors can reasonably expect that their code
will never receive one. They are documented here for completeness, and for the
aid of Sopel's core development.
"""
from __future__ import annotations

from datetime import datetime, timezone
import re
from typing import (
    cast,
    Match,
    Optional,
    Sequence,
    TYPE_CHECKING,
)

from sopel import formatting, tools
from sopel.tools import web
from sopel.tools.identifiers import Identifier, IdentifierFactory

if TYPE_CHECKING:
    from sopel import config


__all__ = [
    'PreTrigger',
    'Trigger',
]

COMMANDS_WITH_CONTEXT = frozenset({
    'INVITE',
    'JOIN',
    'KICK',
    'MODE',
    'NOTICE',
    'PART',
    'PRIVMSG',
    'TOPIC',
})
"""Set of commands with a :attr:`trigger.sender<Trigger.sender>`.

Most IRC messages a plugin will want to handle (channel messages and PMs) are
associated with a context, exposed in the Trigger's ``sender`` property.
However, not *all* commands can be directly associated to a channel or nick.

For IRC message types other than those listed here, the ``trigger``\'s
``sender`` property will be ``None``.
"""


class PreTrigger:
    """A parsed raw message from the server.

    :param str own_nick: the bot's own IRC nickname
    :param str line: the full line from the server
    :param tuple url_schemes: allowed schemes for URL detection

    At the :class:`PreTrigger` stage, the line has not been matched against any
    rules yet. This is what Sopel uses to perform matching.

    ``own_nick`` is needed to correctly parse who sent the line (Sopel or
    someone else).

    ``line`` can also be a simulated echo-message, useful if the connected
    server does not support the capability.

    .. py:attribute:: args

        The IRC command's arguments.

        These are split on spaces, per the IRC protocol.

    .. py:attribute:: ctcp

        The CTCP command name, if present (``None`` otherwise)

    .. py:attribute:: event

        The IRC command name or numeric value.

        See :class:`sopel.tools.events` for a built-in mapping of names to
        numeric values.

    .. py:attribute:: host

        The sender's hostname.

    .. py:attribute:: hostmask

        The sender's hostmask, as sent by the IRC server.

    .. py:attribute:: line

        The raw line received from the server.

    .. py:attribute:: nick

        The nickname that sent this command.

    .. py:attribute:: sender

        Channel name or query where this message was received.

        .. warning::

            The ``sender`` Will be ``None`` for commands that have no implicit
            channel or private-message context, for example AWAY or QUIT.

            The :attr:`COMMANDS_WITH_CONTEXT` attribute lists IRC commands for
            which ``sender`` can be relied upon.

    .. py:attribute:: tags

        Any IRCv3 message tags attached to the line, as a :class:`dict`.

    .. py:attribute:: text

        The last argument of the IRC command.

        If the line contains ``:``, :attr:`text` will be the text following it.

        For lines that do *not* contain ``:``, :attr:`text` will be the last
        argument in :attr:`args` instead.

    .. py:attribute:: urls
        :type: tuple

        List of URLs found in the :attr:`text`.

        This is for ``PRIVMSG`` and ``NOTICE`` messages only. For other
        messages, this will be an empty ``tuple``.

    .. py:attribute:: plain

        The last argument of the IRC command with control codes stripped.

    .. py:attribute:: time

        The time when the message was received.

        If the IRC server sent a message tag indicating when *it* received the
        message, that is used instead of the time when Sopel received it.

        .. versionchanged:: 8.0
            Now a timezone-aware ``datetime`` object.

    .. py:attribute:: user

        The sender's local username.

    """
    component_regex = re.compile(r'([^!]*)!?([^@]*)@?(.*)')
    ctcp_regex = re.compile('\x01(\\S+) ?(.*)\x01')

    def __init__(
        self,
        own_nick: Identifier,
        line: str,
        url_schemes: Optional[Sequence] = None,
        identifier_factory: IdentifierFactory = Identifier,
        statusmsg_prefixes: tuple[str, ...] = tuple(),
    ):
        self.make_identifier: IdentifierFactory = identifier_factory
        line = line.strip('\r\n')
        self.line: str = line
        self.urls: tuple[str, ...] = tuple()
        self.plain: str = ''
        self.ctcp: Optional[str] = None

        # Break off IRCv3 message tags, if present
        self.tags: dict[str, Optional[str]] = {}
        if line.startswith('@'):
            tagstring, line = line.split(' ', 1)
            for raw_tag in tagstring[1:].split(';'):
                tag = raw_tag.split('=', 1)
                if len(tag) > 1:
                    self.tags[tag[0]] = tag[1]
                else:
                    self.tags[tag[0]] = None

        # Client time or server time
        self.time = datetime.now(timezone.utc)
        if 'time' in self.tags:
            # ensure "time" is a string (typecheck)
            tag_time = self.tags['time'] or ''
            try:
                self.time = datetime.strptime(
                    tag_time,
                    "%Y-%m-%dT%H:%M:%S.%fZ",
                ).replace(tzinfo=timezone.utc)
            except ValueError:
                pass  # Server isn't conforming to spec, ignore the server-time

        # Grabs hostmask from line.
        # Example: line = ':Sopel!foo@bar PRIVMSG #sopel :foobar!'
        #          print(hostmask)  # Sopel!foo@bar
        # All lines start with ":" except PING.
        self.hostmask: Optional[str]
        if line.startswith(':'):
            self.hostmask, line = line[1:].split(' ', 1)
        else:
            self.hostmask = None

        # Parses the line into a list of arguments.
        # Some events like MODE don't have a secondary string argument, i.e. no ' :' inside the line.
        # Example 1:  line = ':nick!ident@domain PRIVMSG #sopel :foo bar!'
        #             print(text)    # 'foo bar!'
        #             print(argstr)  # ':nick!ident@domain PRIVMSG #sopel'
        #             print(args)    # [':nick!ident@domain', 'PRIVMSG', '#sopel', 'foo bar!']
        # Example 2:  line = 'irc.libera.chat MODE Sopel +i'
        #             print(text)    # '+i'
        #             print(args)    # ['irc.libera.chat', 'MODE', 'Sopel', '+i']
        if ' :' in line:
            argstr, self.text = line.split(' :', 1)
            self.args = argstr.split(' ')
            self.args.append(self.text)
        else:
            self.args = line.split(' ')
            # see `text` attr documentation above, and #2360
            self.text = self.args[-1] if len(self.args) > 1 else ''

        self.event = self.args[0]
        self.args = self.args[1:]

        # The regex will always match any string, even an empty one
        components_match = cast(
            Match, PreTrigger.component_regex.match(self.hostmask or ''))
        nick, self.user, self.host = components_match.groups()
        self.nick: Identifier = self.make_identifier(nick)

        # If we have arguments, the first one is *usually* the sender,
        # most numerics and certain general events (e.g. QUIT) excepted
        target: Optional[Identifier] = None
        status_prefix: Optional[str] = None

        if self.args and self.event in COMMANDS_WITH_CONTEXT:
            raw_target = self.args[0]
            if statusmsg_prefixes and raw_target[0] in statusmsg_prefixes:
                status_prefix, raw_target = raw_target[0], raw_target[1:]
            target = self.make_identifier(raw_target)

            # Unless we're messaging the bot directly, in which case that
            # second arg will be our bot's name.
            if target.lower() == own_nick.lower():
                target = self.nick

        self.sender = target
        self.status_prefix = status_prefix

        # Parse CTCP
        if self.event == 'PRIVMSG' or self.event == 'NOTICE':
            ctcp_match = PreTrigger.ctcp_regex.match(self.args[-1])
            if ctcp_match is not None:
                ctcp, message = ctcp_match.groups()
                self.ctcp = ctcp
                self.args[-1] = message or ''

            # Search URLs after CTCP parsing
            self.urls = tuple(
                web.search_urls(self.args[-1], schemes=url_schemes))

        # Populate account from extended-join messages
        if self.event == 'JOIN' and len(self.args) == 3:
            # Account is the second arg `...JOIN #Sopel account :realname`
            self.tags['account'] = self.args[1]

        # get plain text message
        if self.args:
            self.plain = formatting.plain(self.args[-1])


class Trigger(str):
    """A line from the server, which has matched a callable's rules.

    :param config: Sopel's current configuration settings object
    :type config: :class:`~sopel.config.Config`
    :param message: the message that matched the callable
    :type message: :class:`PreTrigger`
    :param match: what *in* the message matched
    :type match: :ref:`Match object <match-objects>`
    :param str account: services account name of the ``message``'s sender
                        (optional; only applies on networks with the
                        ``account-tag`` capability enabled)

    A :class:`Trigger` object itself can be used as a string; when used in
    this way, it represents the matching line's full text.

    The ``match`` is based on the matching regular expression rule; Sopel's
    command decorators auto-generate rules containing specific numbered groups
    that are documented separately. (See :attr:`group` below.)

    Note that CTCP messages (``PRIVMSG``\\es and ``NOTICE``\\es which start
    and end with ``'\\x01'``) will have the ``'\\x01'`` bytes stripped, and
    :attr:`trigger.ctcp <ctcp>` will contain the command (e.g. ``ACTION``).

    .. note::

        CTCP used to be stored as the ``intent`` tag. Since message intents
        never made it past the IRCv3 draft stage, Sopel dropped support for
        them in Sopel 8.

    """
    sender = property(lambda self: self._pretrigger.sender)
    """Where the message arrived from.

    :type: :class:`~sopel.tools.identifiers.Identifier`

    This will be a channel name for "regular" (channel) messages, or the nick
    that sent a private message.

    You can check if the trigger comes from a channel or a nick with its
    :meth:`~sopel.tools.identifiers.Identifier.is_nick` method::

        if trigger.sender.is_nick():
            # message sent from a private message
        else:
            # message sent from a channel

    .. important::

        If the message was sent to a `specific status prefix`__, the ``sender``
        does not include the status prefix. Be sure to use the
        :attr:`status_prefix` when replying.

        Note that the ``bot`` argument passed to plugin callables is a
        :class:`~sopel.bot.SopelWrapper` that handles this for the default
        ``destination`` of the methods it overrides (most importantly,
        :meth:`~sopel.bot.SopelWrapper.say` &
        :meth:`~sopel.bot.SopelWrapper.reply`).

    .. warning::

        The ``sender`` Will be ``None`` for commands that have no implicit
        channel or private-message context, for example AWAY or QUIT.

        The :attr:`COMMANDS_WITH_CONTEXT` attribute lists IRC commands for
        which ``sender`` can be relied upon.

    .. __: https://modern.ircdocs.horse/#statusmsg-parameter
    """
    status_prefix = property(lambda self: self._pretrigger.status_prefix)
    """The prefix used for the :attr:`sender` for status-specific messages.

    :type: Optional[str]

    This will be ``None`` by default. If a message is sent to a channel, it may
    have a status prefix for status-specific messages. In that case, the prefix
    is removed from the channel's identifier (see :attr:`sender`) and saved to
    this attribute.
    """
    time = property(lambda self: self._pretrigger.time)
    """When the message was received.

    :type: naïve :class:`~datetime.datetime` object (no timezone)

    If the IRC server supports ``server-time``, :attr:`time` will give that
    value. Otherwise, :attr:`time` will use the time when the message was
    received by Sopel. In both cases, this time is in UTC.
    """
    raw = property(lambda self: self._pretrigger.line)
    """The entire raw IRC message, as sent from the server.

    :type: str

    This includes the CTCP ``\\x01`` bytes and command, if they were included.
    """
    is_privmsg = property(lambda self: self._is_privmsg)
    """Whether the message was triggered in a private message.

    :type: bool

    ``True`` if the trigger was received in a private message; ``False`` if it
    came from a channel.
    """
    hostmask = property(lambda self: self._pretrigger.hostmask)
    """Hostmask of the person who sent the message as ``<nick>!<user>@<host>``.

    :type: str
    """
    user = property(lambda self: self._pretrigger.user)
    """Local username of the person who sent the message.

    :type: str
    """
    nick = property(lambda self: self._pretrigger.nick)
    """The nickname who sent the message.

    :type: :class:`~sopel.tools.identifiers.Identifier`
    """
    host = property(lambda self: self._pretrigger.host)
    """The hostname of the person who sent the message.

    :type: str
    """
    event = property(lambda self: self._pretrigger.event)
    """The IRC event which triggered the message.

    :type: str

    Most plugin :func:`callables <callable>` primarily need to deal with
    ``PRIVMSG``. Other event types like ``NOTICE``, ``NICK``, ``TOPIC``,
    ``KICK``, etc. must be requested using :func:`.plugin.event`.
    """
    ctcp = property(lambda self: self._pretrigger.ctcp)
    """The CTCP command (if any).

    :type: str

    Common CTCP commands are ``ACTION``, ``VERSION``, and ``TIME``. Other
    commands include ``USERINFO``, ``PING``, and various ``DCC`` operations.

    .. versionadded:: 7.1

    .. important::

        Use this attribute instead of the ``intent`` tag in :attr:`tags`.
        Message intents never made it past the IRCv3 draft stage, and Sopel
        dropped support for them in Sopel 8.

    """
    match = property(lambda self: self._match)
    """The :ref:`Match object <match-objects>` for the triggering line.

    :type: :ref:`Match object <match-objects>`
    """
    group = property(lambda self: self._match.group)
    """The ``group()`` function of the :attr:`match` attribute.

    :type: :term:`method`
    :rtype: str

    Any regular-expression :func:`rules <.plugin.rule>` attached to the
    triggered :func:`callable` may define numbered or named groups that can be
    retrieved through this property.

    Sopel's command decorators each define a predetermined set of numbered
    groups containing fragments of the line that plugins commonly use.

    .. seealso::

       For more information on predefined numbered match groups in commands,
       see :func:`.plugin.command`, :func:`.plugin.action_command`, and
       :func:`.plugin.nickname_command`.

       Also see Python's :meth:`re.Match.group` documentation for details
       about this method's behavior.

    """
    groups = property(lambda self: self._match.groups)
    """The ``groups()`` function of the :attr:`match` attribute.

    :type: :term:`method`
    :rtype: tuple

    See Python's :meth:`re.Match.groups` documentation for details.
    """
    groupdict = property(lambda self: self._match.groupdict)
    """The ``groupdict()`` function of the :attr:`match` attribute.

    :type: :term:`method`
    :rtype: dict

    See Python's :meth:`re.Match.groupdict` documentation for details.
    """
    args = property(lambda self: self._pretrigger.args)
    """A list containing each of the arguments to an event.

    :type: list[str]

    These are the strings passed between the event name and the colon. For
    example, when setting ``mode -m`` on the channel ``#example``, args would
    be ``['#example', '-m']``
    """
    urls = property(lambda self: self._pretrigger.urls)
    """A tuple containing all URLs found in the text.

    :type: tuple

    URLs are listed only for ``PRIVMSG`` or a ``NOTICE``, otherwise this is
    an empty tuple.
    """
    plain = property(lambda self: self._pretrigger.plain)
    """The text without formatting control codes.

    :type: str

    This is the text of the trigger object without formatting control codes.
    """
    tags = property(lambda self: self._pretrigger.tags)
    """A map of the IRCv3 message tags on the message.

    :type: dict
    """
    admin = property(lambda self: self._admin)
    """Whether the triggering :attr:`nick` is one of the bot's admins.

    :type: bool

    ``True`` if the triggering :attr:`nick` is a Sopel admin; ``False`` if not.

    Note that Sopel's :attr:`~.config.core_section.CoreSection.owner` is also
    considered to be an admin.
    """
    owner = property(lambda self: self._owner)
    """Whether the :attr:`nick` which triggered the command is the bot's owner.

    :type: bool

    ``True`` if the triggering :attr:`nick` is Sopel's owner; ``False`` if not.
    """
    account = property(lambda self: self.tags.get('account') or self._account)
    """The services account name of the user sending the message.

    :type: str or None

    Note: This is only available if the ``account-tag`` capability or *both*
    the ``account-notify`` and ``extended-join`` capabilities are available on
    the connected IRC network. If this is not the case, or if the user sending
    the message isn't logged in to services, this property will be ``None``.
    """

    def __new__(
        cls,
        settings: config.Config,
        message: PreTrigger,
        match: Match,
        account: Optional[str] = None,
    ) -> 'Trigger':
        return str.__new__(cls, message.args[-1] if message.args else '')

    def __init__(
        self,
        settings: config.Config,
        message: PreTrigger,
        match: Match,
        account: Optional[str] = None,
    ) -> None:
        self._account = account
        self._pretrigger = message
        self._match = match
        self._is_privmsg = message.sender and message.sender.is_nick()

        def match_host_or_nick(pattern):
            pattern = tools.get_hostmask_regex(pattern)
            return bool(
                pattern.match(self.nick) or
                pattern.match('@'.join((self.nick, self.host)))
            )

        if settings.core.owner_account:
            self._owner = settings.core.owner_account == self.account
        else:
            self._owner = match_host_or_nick(settings.core.owner)
        self._admin = (
            self._owner or
            self.account in settings.core.admin_accounts or
            any(match_host_or_nick(item) for item in settings.core.admins)
        )



"""Useful miscellaneous tools and shortcuts for Sopel plugins

.. versionadded:: 3.0
"""

# tools.py - Sopel misc tools
# Copyright 2008, Sean B. Palmer, inamidst.com
# Copyright © 2012, Elad Alfassa <elad@fedoraproject.org>
# Copyright 2012, Elsie Powell, embolalia.com
# Licensed under the Eiffel Forum License 2.

# https://sopel.chat

from __future__ import annotations

import codecs
import logging
import re
import sys

from sopel.lifecycle import deprecated  # Don't delete; maintains backward compatibility with pre-8.0 API
from ._events import events  # NOQA

# shortcuts & backward compatibility with pre-8.0
from .identifiers import Identifier  # NOQA
from .memories import (  # NOQA
    SopelIdentifierMemory,
    SopelMemory,
    SopelMemoryWithDefault,
)
from . import time, web  # NOQA


# Long kept for Python compatibility, but it's time we let these go.
raw_input = deprecated(  # pragma: no cover
    'Use the `input` function directly.',
    version='8.0',
    removed_in='8.1',
    func=input)
iteritems = deprecated(  # pragma: no cover
    "Use the dict object's `.items()` method directly.",
    version='8.0',
    removed_in='8.1',
    func=dict.items)
itervalues = deprecated(  # pragma: no cover
    "Use the dict object's `.values()` method directly.",
    version='8.0',
    removed_in='8.1',
    func=dict.values)
iterkeys = deprecated(  # pragma: no cover
    "Use the dict object's `.keys()` method directly.",
    version='8.0',
    removed_in='8.1',
    func=dict.keys)


@deprecated('Shim for Python 2 cross-compatibility, no longer needed. '
            'Use built-in `input()` instead.',
            version='7.1',
            warning_in='8.0',
            removed_in='8.1')
def get_input(prompt):
    """Get decoded input from the terminal (equivalent to Python 3's ``input``).

    :param str prompt: what to display as a prompt on the terminal
    :return: the user's input
    :rtype: str

    .. deprecated:: 7.1

        This function will be removed in Sopel 8.1.

    """
    return input(prompt)


def get_sendable_message(text, max_length=400):
    """Get a sendable ``text`` message, with its excess when needed.

    :param str txt: text to send (expects Unicode-encoded string)
    :param int max_length: maximum length of the message to be sendable
    :return: a tuple of two values, the sendable text and its excess text
    :rtype: (str, str)

    We're arbitrarily saying that the (default) max is 400 bytes of text
    when messages will be split, but callers can specify a different value
    (e.g. to account precisely for the bot's hostmask).

    The ``max_length`` is the max length of text in **bytes**, but we take
    care of multibyte UTF-8 characters by working on the Unicode string,
    then making sure the bytes version is smaller than the max length.

    .. note::

        In most cases, letting the bot gracefully handle message truncation
        using optional arguments to :meth:`bot.say() <.bot.Sopel.say>` is
        preferable. However, this function is part of the public API to provide
        for more advanced use-cases.

        See also the :meth:`bot.safe_text_length() <.bot.Sopel.safe_text_length>`
        method, whose return value can be passed as this function's
        ``max_length`` argument.

    .. versionadded:: 6.6.2
    """
    unicode_max_length = max_length
    excess = ''

    while len(text.encode('utf-8')) > max_length:
        last_space = text.rfind(' ', 0, unicode_max_length)
        if last_space == -1:
            # No last space, just split where it is possible
            excess = text[unicode_max_length:] + excess
            text = text[:unicode_max_length]
            # Decrease max length for the unicode string
            unicode_max_length = unicode_max_length - 1
        else:
            # Split at the last best space found
            excess = text[last_space:] + excess
            text = text[:last_space]

    return text, excess.lstrip()


class OutputRedirect:
    """Redirect the output to the terminal and a log file.

    A simplified object used to write to both the terminal and a log file.

    .. deprecated:: 8.0

        Vestige of old logging system. Will be removed in Sopel 8.1.

    """

    @deprecated(
        "Remnant of Sopel's pre-7.0 logging system. "
        "You shouldn't have been using this anyway.",
        version='8.0',
        removed_in='8.1',
    )
    def __init__(self, logpath, stderr=False, quiet=False):
        """Create an object which will log to both a file and the terminal.

        :param str logpath: path to the log file
        :param bool stderr: write output to stderr if ``True``, or to stdout
                            otherwise
        :param bool quiet: write to the log file only if ``True`` (and not to
                           the terminal)

        Create an object which will log to the file at ``logpath`` as well as
        the terminal.
        """
        self.logpath = logpath
        self.stderr = stderr
        self.quiet = quiet

    def write(self, string):
        """Write the given ``string`` to the logfile and terminal.

        :param str string: the string to write
        """
        if not self.quiet and sys.__stderr__ and sys.__stdout__:
            try:
                if self.stderr:
                    sys.__stderr__.write(string)
                else:
                    sys.__stdout__.write(string)
            except Exception:  # TODO: Be specific
                pass

        with codecs.open(self.logpath, 'ab', encoding="utf8",
                         errors='xmlcharrefreplace') as logfile:
            try:
                logfile.write(string)
            except UnicodeDecodeError:
                # we got an invalid string, safely encode it to utf-8
                logfile.write(str(string, 'utf8', errors="replace"))

    def flush(self):
        """Flush the file writing buffer."""
        if self.stderr and sys.__stderr__:
            sys.__stderr__.flush()
        elif sys.__stdout__:
            sys.__stdout__.flush()


@deprecated(
    "Plugins should use a logger object.",
    version='8.0',
    removed_in='8.1',
)
def stderr(string):
    # don't like importing inside functions, but circular imports ensue otherwise
    from sopel.cli.utils import stderr as _stderr

    return _stderr(string)


@deprecated(
    "Function for internal use, moved to `sopel.cli.utils`.",
    version='8.0',
    removed_in='8.1',
)
def check_pid(pid):
    # don't like importing inside functions, but circular imports ensue otherwise
    from sopel.cli.utils import check_pid as _check_pid

    return _check_pid(pid)


def get_hostmask_regex(mask):
    """Get a compiled regex pattern for an IRC hostmask

    :param str mask: the hostmask that the pattern should match
    :return: a compiled regex pattern matching the given ``mask``
    :rtype: :ref:`re.Pattern <python:re-objects>`

    .. versionadded:: 4.4
    """
    mask = re.escape(mask)
    mask = mask.replace(r'\*', '.*')
    return re.compile(mask + '$', re.I)


def get_logger(plugin_name):
    """Return a logger for a plugin.

    :param str plugin_name: name of the plugin
    :return: the logger for the given plugin

    This::

        from sopel import tools
        LOGGER = tools.get_logger('my_custom_plugin')

    is equivalent to this::

        import logging
        LOGGER = logging.getLogger('sopel.externals.my_custom_plugin')

    Internally, Sopel configures logging for the ``sopel`` namespace, so
    external plugins can't benefit from it with ``logging.getLogger(__name__)``
    as they won't be in the same namespace. This function uses the
    ``plugin_name`` with a prefix inside this namespace.

    .. versionadded:: 7.0
    """
    return logging.getLogger('sopel.externals.%s' % plugin_name)


def chain_loaders(*lazy_loaders):
    """Chain lazy loaders into one.

    :param lazy_loaders: one or more lazy loader functions
    :type lazy_loaders: :term:`function`
    :return: a lazy loader that combines all of the given ones
    :rtype: :term:`function`

    This function takes any number of lazy loaders as arguments and merges them
    together into one. It's primarily a helper for lazy rule decorators such as
    :func:`sopel.plugin.url_lazy`.

    .. versionadded:: 7.1

    .. important::

        This function doesn't check the uniqueness of regexes generated by
        all the loaders.

    """
    def chained_loader(settings):
        return [
            regex
            for lazy_loader in lazy_loaders
            for regex in lazy_loader(settings)
        ]
    return chained_loader



"""
admin.py - Sopel Admin Plugin
Copyright 2010-2011, Sean B. Palmer (inamidst.com) and Michael Yanovich
(yanovich.net)
Copyright © 2012, Elad Alfassa, <elad@fedoraproject.org>
Copyright 2013, Ari Koivula <ari@koivu.la>
Copyright 2019, Florian Strzelecki, https://github.com/Exirel
Licensed under the Eiffel Forum License 2.

https://sopel.chat
"""
from __future__ import annotations

import logging

from sopel import plugin
from sopel.config import types

LOGGER = logging.getLogger(__name__)

ERROR_JOIN_NO_CHANNEL = 'Which channel should I join?'
"""Error message when channel is missing from command arguments."""
ERROR_PART_NO_CHANNEL = 'Which channel should I quit?'
"""Error message when channel is missing from command arguments."""
ERROR_NOTHING_TO_SAY = 'I need a channel and a message to talk.'
"""Error message when channel and/or message are missing."""
ERROR_NOTHING_TO_RAW = 'I need an IRC message to send.'
"""Error message when no raw IRC message was given."""


class AdminSection(types.StaticSection):
    hold_ground = types.BooleanAttribute('hold_ground', default=False)
    """Auto re-join on kick"""
    auto_accept_invite = types.BooleanAttribute('auto_accept_invite', default=True)
    """Auto-join channels when invited"""


def configure(config):
    """
    | name | example | purpose |
    | ---- | ------- | ------- |
    | hold\\_ground | False | Auto-rejoin the channel after being kicked. |
    | auto\\_accept\\_invite | True | Auto-join channels when invited. |
    """
    config.define_section('admin', AdminSection)
    config.admin.configure_setting('hold_ground',
                                   "Automatically re-join after being kicked?")
    config.admin.configure_setting('auto_accept_invite',
                                   'Automatically join channels when invited?')


def setup(bot):
    bot.config.define_section('admin', AdminSection)


class InvalidSection(Exception):
    def __init__(self, section):
        super().__init__(self, 'Section [{}] does not exist.'.format(section))
        self.section = section


class InvalidSectionOption(Exception):
    def __init__(self, section, option):
        super().__init__(self, 'Section [{}] does not have option \'{}\'.'.format(section, option))
        self.section = section
        self.option = option


def _get_config_channels(channels):
    """List"""
    for channel_info in channels:
        if ' ' in channel_info:
            yield channel_info.split(' ', 1)
        else:
            yield (channel_info, None)


def _set_config_channels(bot, channels):
    bot.config.core.channels = [
        ' '.join([part for part in items if part])
        for items in channels.items()
    ]
    bot.config.save()


def _join(bot, channel, key=None, save=True):
    if not channel:
        return

    if not key:
        bot.join(channel)
    else:
        bot.join(channel, key)

    if save:
        channels = dict(_get_config_channels(bot.config.core.channels))
        # save only if channel is new or key has been changed
        if channel not in channels or channels[channel] != key:
            channels[channel] = key
            _set_config_channels(bot, channels)
            LOGGER.info('Added "%s" to core.channels.', channel)


def _part(bot, channel, msg=None, save=True):
    bot.part(channel, msg or None)

    if save:
        channels = dict(_get_config_channels(bot.config.core.channels))
        if channel in channels:
            del channels[channel]
            _set_config_channels(bot, channels)
            LOGGER.info('Removed "%s" from core.channels.', channel)


@plugin.require_privmsg
@plugin.require_admin
@plugin.command('join')
@plugin.priority('low')
@plugin.example('.join #example key', user_help=True)
@plugin.example('.join #example', user_help=True)
def join(bot, trigger):
    """Join the specified channel. This is an admin-only command."""
    channel, key = trigger.group(3), trigger.group(4)
    if not channel:
        bot.reply(ERROR_JOIN_NO_CHANNEL)
        return

    _join(bot, channel, key)


@plugin.require_privmsg
@plugin.require_admin
@plugin.command('tmpjoin')
@plugin.priority('low')
@plugin.example('.tmpjoin #example key', user_help=True)
@plugin.example('.tmpjoin #example', user_help=True)
def temporary_join(bot, trigger):
    """Like ``join``, without saving. This is an admin-only command.

    Unlike the ``join`` command, ``tmpjoin`` won't remember the channel upon
    restarting the bot.
    """
    channel, key = trigger.group(3), trigger.group(4)
    if not channel:
        bot.reply(ERROR_JOIN_NO_CHANNEL)
        return

    _join(bot, channel, key, save=False)


@plugin.require_privmsg
@plugin.require_admin
@plugin.command('part')
@plugin.priority('low')
@plugin.example('.part #example')
def part(bot, trigger):
    """Part the specified channel. This is an admin-only command."""
    channel, _sep, part_msg = trigger.group(2).partition(' ')
    if not channel:
        bot.reply(ERROR_PART_NO_CHANNEL)
        return

    _part(bot, channel, part_msg)


@plugin.require_privmsg
@plugin.require_admin
@plugin.command('tmppart')
@plugin.priority('low')
@plugin.example('.tmppart #example')
def temporary_part(bot, trigger):
    """Like ``part``, without saving. This is an admin-only command.

    Unlike the ``part`` command, ``tmppart`` will rejoin the channel upon
    restarting the bot.
    """
    channel, _sep, part_msg = trigger.group(2).partition(' ')
    if not channel:
        bot.reply(ERROR_PART_NO_CHANNEL)
        return

    _part(bot, channel, part_msg, save=False)


@plugin.require_privmsg
@plugin.require_admin
@plugin.command('chanlist', 'channels')
@plugin.priority('low')
def channel_list(bot, trigger):
    """Show channels Sopel is in."""
    channels = ', '.join(sorted(bot.channels.keys()))

    # conservative assumption about how much room we have in the line to make
    # sure `max_messages` won't actually truncate anything
    bot.say(channels, max_messages=1 + len(channels) // 400)


@plugin.require_privmsg
@plugin.require_owner
@plugin.command('restart')
@plugin.priority('low')
def restart(bot, trigger):
    """Restart the bot. This is an owner-only command."""
    quit_message = trigger.group(2)
    default_message = 'Restarting on command from %s.' % trigger.nick
    if not quit_message:
        quit_message = default_message

    LOGGER.info(default_message)
    bot.restart(quit_message)


@plugin.require_privmsg
@plugin.require_owner
@plugin.command('quit')
@plugin.priority('low')
def quit(bot, trigger):
    """Quit from the server. This is an owner-only command."""
    quit_message = trigger.group(2)
    default_message = 'Quitting on command from %s.' % trigger.nick
    if not quit_message:
        quit_message = default_message

    LOGGER.info(default_message)
    bot.quit(quit_message)


@plugin.require_privmsg
@plugin.require_owner
@plugin.command('raw')
@plugin.priority('low')
@plugin.example('.raw PRIVMSG NickServ :CERT ADD')
def raw(bot, trigger):
    """
    Send a raw IRC message. Can only be done in privmsg by the bot owner.
    This is mostly useful for debugging.
    """
    if trigger.group(2) is None:
        bot.reply(ERROR_NOTHING_TO_RAW)
        return

    bot.write([trigger.group(2)])


@plugin.require_privmsg
@plugin.require_admin
@plugin.command('say', 'msg')
@plugin.priority('low')
@plugin.example('.say #YourPants Does anyone else smell neurotoxin?')
def say(bot, trigger):
    """
    Send a message to a given channel or nick. Can only be done in privmsg by
    an admin.
    """
    if trigger.group(2) is None:
        bot.reply(ERROR_NOTHING_TO_SAY)
        return

    channel, _sep, message = trigger.group(2).partition(' ')
    message = message.strip()
    if not channel or not message:
        bot.reply(ERROR_NOTHING_TO_SAY)
        return

    bot.say(message, channel)


@plugin.require_privmsg
@plugin.require_admin
@plugin.command('me')
@plugin.priority('low')
def me(bot, trigger):
    """
    Send an ACTION (/me) to a given channel or nick. Can only be done in
    privmsg by an admin.
    """
    if trigger.group(2) is None:
        bot.reply(ERROR_NOTHING_TO_SAY)
        return

    channel, _sep, action = trigger.group(2).partition(' ')
    action = action.strip()
    if not channel or not action:
        bot.reply(ERROR_NOTHING_TO_SAY)
        return

    bot.action(action, channel)


@plugin.event('INVITE')
@plugin.priority('low')
def invite_join(bot, trigger):
    """Join a channel Sopel is invited to, if the inviter is an admin."""
    channel = trigger.args[1]
    if trigger.admin:
        LOGGER.info(
            'Got invited to "%s" by an admin.', channel)
        bot.join(channel)
    elif bot.config.admin.auto_accept_invite:
        LOGGER.info(
            'Got invited to "%s"; admin.auto_accept_invite is on', channel)
        bot.join(channel)
    else:
        LOGGER.info(
            'Got invited to "%s"; admin.auto_accept_invite is off.', channel)


@plugin.event('KICK')
@plugin.priority('low')
def hold_ground(bot, trigger):
    """
    This function monitors all kicks across all channels Sopel is in. If it
    detects that it is the one kicked it'll automatically join that channel.

    WARNING: This may not be needed and could cause problems if Sopel becomes
    annoying. Please use this with caution.
    """
    if bot.nick != trigger.args[1]:
        # not the bot; ignore
        return

    channel = trigger.sender
    if bot.config.admin.hold_ground:
        LOGGER.info('Got kicked from "%s"; admin.hold_ground is on.', channel)
        bot.join(channel)
    else:
        LOGGER.info('Got kicked from "%s"; admin.hold_ground is off.', channel)


@plugin.require_privmsg
@plugin.require_admin
@plugin.command('mode')
@plugin.priority('low')
def mode(bot, trigger):
    """Set a user mode on Sopel. Can only be done in privmsg by an admin."""
    mode = trigger.group(3)
    if not mode:
        bot.reply('What mode should I set?')

    bot.write(('MODE', bot.nick, mode))


def parse_section_option_value(config, trigger):
    """Parse trigger for set/unset to get relevant config elements.

    :param config: Sopel's config
    :param trigger: IRC line trigger
    :return: A tuple with ``(section, section_name, static_sec, option, value)``
    :raises InvalidSection: section does not exist
    :raises InvalidSectionOption: option does not exist for section

    The ``value`` is optional and can be returned as ``None`` if omitted from command.
    """
    match = trigger.group(3)
    if match is None:
        raise ValueError  # Invalid command

    # Get section and option from first argument.
    arg1 = match.split('.')
    if len(arg1) == 1:
        section_name, option = "core", arg1[0]
    elif len(arg1) == 2:
        section_name, option = arg1
    else:
        raise ValueError  # invalid command format

    section = getattr(config, section_name, False)
    if not section:
        raise InvalidSection(section_name)
    static_sec = isinstance(section, types.StaticSection)

    if static_sec and not hasattr(section, option):
        raise InvalidSectionOption(section_name, option)  # Option not found in section

    if not static_sec and not config.parser.has_option(section_name, option):
        raise InvalidSectionOption(section_name, option)  # Option not found in section

    delim = trigger.group(2).find(' ')
    # Skip preceding whitespaces, if any.
    while delim > 0 and delim < len(trigger.group(2)) and trigger.group(2)[delim] == ' ':
        delim = delim + 1

    value = trigger.group(2)[delim:]
    if delim == -1 or delim == len(trigger.group(2)):
        value = None

    return (section, section_name, static_sec, option, value)


@plugin.require_privmsg("This command only works as a private message.")
@plugin.require_admin("This command requires admin privileges.")
@plugin.command('set')
@plugin.example('.set core.owner MyNick')
def set_config(bot, trigger):
    """See and modify values of Sopel's config object.

    Trigger args:
        arg1 - section and option, in the form "section.option"
        arg2 - value

    If there is no section, section will default to "core".
    If value is not provided, the current value will be displayed.
    """
    try:
        section, section_name, static_sec, option, value = parse_section_option_value(bot.config, trigger)
    except ValueError:
        bot.say('Usage: {}set section.option [value]'.format(bot.config.core.help_prefix))
        return
    except (InvalidSection, InvalidSectionOption) as exc:
        bot.say(exc.args[1])
        return

    # Get a descriptor class for the option if it's a static section
    descriptor = getattr(section.__class__, option) if static_sec else None

    # Display current value if no value is given
    if not value:
        value = getattr(section, option)

        if descriptor is not None:
            if getattr(descriptor, 'is_secret', False):
                # Keep secret option as secret
                value = "(secret value censored)"
        elif option.endswith("password") or option.endswith("pass"):
            # Fallback to guessing if secret, for backward compatiblity
            # TODO: consider a deprecation warning when loading settings
            value = "(password censored)"

        bot.say("%s.%s = %s (%s)" % (section_name, option, value, type(value).__name__))
        return

    # Owner-related settings cannot be modified interactively. Any changes to these
    # settings must be made directly in the config file.
    if section_name == 'core' and option in ['owner', 'owner_account']:
        bot.say("Changing '{}.{}' requires manually editing the configuration file."
                .format(section_name, option))
        return

    # Otherwise, set the value to one given
    if descriptor is not None:
        try:
            value = descriptor._parse(value, bot.config, section)
        except ValueError as exc:
            bot.say("Can't set attribute: " + str(exc))
            return
    setattr(section, option, value)
    LOGGER.info('%s.%s set successfully.', section_name, option)
    bot.say("OK. Set '{}.{}' successfully.".format(section_name, option))


@plugin.require_privmsg("This command only works as a private message.")
@plugin.require_admin("This command requires admin privileges.")
@plugin.command('unset')
@plugin.example('.unset core.owner')
def unset_config(bot, trigger):
    """Unset value of Sopel's config object.

    Unsetting a value will reset it to the default specified in the config
    definition.

    Trigger args:
        arg1 - section and option, in the form "section.option"

    If there is no section, section will default to "core".
    """
    try:
        section, section_name, static_sec, option, value = parse_section_option_value(bot.config, trigger)
    except ValueError:
        bot.say('Usage: {}unset section.option [value]'.format(bot.config.core.help_prefix))
        return
    except (InvalidSection, InvalidSectionOption) as exc:
        bot.say(exc.args[1])
        return

    if value:
        bot.say('Invalid command; no value should be provided to unset.')
        return

    try:
        setattr(section, option, None)
        LOGGER.info('%s.%s unset.', section_name, option)
        bot.say("Unset '{}.{}' successfully.".format(section_name, option))
    except ValueError:
        bot.say('Cannot unset {}.{}; it is a required option.'.format(section_name, option))


@plugin.require_privmsg
@plugin.require_admin
@plugin.command('save')
@plugin.example('.save')
def save_config(bot, trigger):
    """Save state of Sopel's config object to the configuration file."""
    bot.config.save()
    LOGGER.info('Configuration file saved.')
    bot.say('Configuration file saved.')



"""
pronouns.py - Sopel Pronouns Plugin
Copyright © 2016, Elsie Powell
Licensed under the Eiffel Forum License 2.

https://sopel.chat
"""
from __future__ import annotations

import logging

import requests

from sopel import plugin
from sopel.config import types


LOGGER = logging.getLogger(__name__)


class PronounsSection(types.StaticSection):
    fetch_complete_list = types.BooleanAttribute('fetch_complete_list', default=True)
    """Whether to attempt fetching the complete list the web backend uses, at bot startup."""

    link_base_url = types.ValidatedAttribute(
        'link_base_url', default='https://pronouns.sopel.chat')
    """Base URL for links to pronoun info.

    Defaults to an instance of https://github.com/sopel-irc/pronoun-service
    hosted by the Sopel project, but it's easy to make your own instance
    available at a custom domain using one of many free static site hosts.
    """


def configure(settings):
    """
    | name | example | purpose |
    | ---- | ------- | ------- |
    | fetch_complete_list | True | Whether to attempt fetching the complete pronoun list from the web backend at startup. |
    | link_base_url | https://pronouns.sopel.chat | Base URL for pronoun info links. |
    """
    settings.define_section('pronouns', PronounsSection)
    settings.pronouns.configure_setting(
        'fetch_complete_list',
        'Fetch the most current list of pronoun sets at startup?')
    settings.pronouns.configure_setting(
        'link_base_url',
        'Base URL for pronoun info links:')


def setup(bot):
    bot.settings.define_section('pronouns', PronounsSection)

    # Copied from svelte-pronounisland, leaving a *lot* out.
    # If ambiguous, the earlier one will be used.
    # This basic set is hard-coded to guarantee that the ten most(ish) common sets
    # will work, even if fetching the current set from GitHub fails.
    bot.memory['pronoun_sets'] = {
        'ze/hir': 'ze/hir/hir/hirs/hirself',
        'ze/zir': 'ze/zir/zir/zirs/zirself',
        'they/.../themselves': 'they/them/their/theirs/themselves',
        'they/.../themself': 'they/them/their/theirs/themself',
        'she/her': 'she/her/her/hers/herself',
        'he/him': 'he/him/his/his/himself',
        'xey/xem': 'xey/xem/xyr/xyrs/xemself',
        'sie/hir': 'sie/hir/hir/hirs/hirself',
        'it/it': 'it/it/its/its/itself',
        'ey/em': 'ey/em/eir/eirs/eirself',
    }

    if not bot.settings.pronouns.fetch_complete_list:
        return

    # and now try to get the current list our fork of the backend uses
    # (https://github.com/sopel-irc/pronoun-service)
    try:
        r = requests.get(
            'https://github.com/sopel-irc/pronoun-service/raw/main/src/lib/data/pronouns.tab')
        r.raise_for_status()
        fetched_pairs = _process_pronoun_sets(r.text.splitlines())
    except requests.exceptions.RequestException:
        # don't do anything, just log the failure and use the hard-coded set
        LOGGER.exception("Couldn't fetch full pronouns list; using default set.")
        return
    except Exception:
        # don't care what failed, honestly, since we aren't trying to fix it
        LOGGER.exception("Couldn't parse fetched pronouns; using default set.")
        return
    else:
        bot.memory['pronoun_sets'] = dict(fetched_pairs)


def _process_pronoun_sets(set_list):
    trie = PronounTrie()
    trie.insert_list(set_list)
    yield from trie.get_pairs()


class PronounTrieNode:
    def __init__(self, source=''):
        self.children = {}
        """Child nodes are stored here."""

        self.freq = 0
        """Store how many times this node is visited during insertion."""

        self.source = source
        """The full pronoun set that caused this node's creation."""


class PronounTrie:
    def __init__(self):
        self.root = PronounTrieNode()
        """A Trie needs a root entry."""

    def insert(self, pronoun_set):
        """Insert a single pronoun set."""
        pronoun_set = pronoun_set.replace('\t', '/')
        current_node = self.root
        for pronoun in pronoun_set.split('/'):
            # create a new node if the path doesn't exist
            # and use it as the current node
            current_node = current_node.children.setdefault(pronoun, PronounTrieNode(pronoun_set))

            # increment frequency
            current_node.freq += 1

    def insert_list(self, set_list):
        """Load a list of pronoun sets all at once."""
        for item in set_list:
            self.insert(item)

    def get_pairs(self, root=None, prefix=''):
        """Yield tuples of ``(prefix, full/pronoun/set)``."""
        if root is None:
            root = self.root

        if root.freq == 1:
            yield prefix, root.source
        else:
            if prefix:
                prefix += '/'
            for word, node in root.children.items():
                yield from self.get_pairs(node, prefix + word)


@plugin.command('pronouns')
@plugin.example('.pronouns Embolalia')
def pronouns(bot, trigger):
    """Show the pronouns for a given user, defaulting to the current user if left blank."""
    if not trigger.group(3):
        pronouns = bot.db.get_nick_value(trigger.nick, 'pronouns')
        if pronouns:
            say_pronouns(bot, trigger.nick, pronouns)
        else:
            bot.reply("I don't know your pronouns! You can set them with "
                      "{}setpronouns".format(bot.settings.core.help_prefix))
    else:
        pronouns = bot.db.get_nick_value(trigger.group(3), 'pronouns')
        if pronouns:
            say_pronouns(bot, trigger.group(3), pronouns)
        elif trigger.group(3) == bot.nick:
            # You can stuff an entry into the database manually for your bot's
            # gender, but like… it's a bot.
            bot.say(
                "I am a bot. Beep boop. My pronouns are it/it/its/its/itself. "
                "See {}/it for examples.".format(bot.settings.pronouns.link_base_url)
            )
        else:
            bot.reply("I don't know {}'s pronouns. They can set them with "
                      "{}setpronouns".format(trigger.group(3),
                                             bot.settings.core.help_prefix))


def say_pronouns(bot, nick, pronouns):
    for short, set_ in bot.memory['pronoun_sets'].items():
        if pronouns == set_:
            break
        short = pronouns

    bot.say(
        "{nick}'s pronouns are {pronouns}. See {base_url}/{short} for examples."
        .format(
            nick=nick,
            pronouns=pronouns,
            base_url=bot.settings.pronouns.link_base_url,
            short=short,
        )
    )


@plugin.command('setpronouns')
@plugin.example('.setpronouns fae/faer/faer/faers/faerself')
@plugin.example('.setpronouns they/them/theirs')
@plugin.example('.setpronouns they/them')
def set_pronouns(bot, trigger):
    """Set your pronouns."""
    requested_pronouns = trigger.group(2)
    if not requested_pronouns:
        bot.reply('What pronouns do you use?')
        return

    disambig = ''
    requested_pronouns_split = requested_pronouns.split("/")
    if len(requested_pronouns_split) < 5:
        matching = []
        for known_pronoun_set in bot.memory['pronoun_sets'].values():
            known_split_set = known_pronoun_set.split("/")
            if known_pronoun_set.startswith(requested_pronouns + "/") or (
                len(requested_pronouns_split) == 3
                and (
                    (
                        # "they/.../themself"
                        requested_pronouns_split[1] == "..."
                        and requested_pronouns_split[0] == known_split_set[0]
                        and requested_pronouns_split[2] == known_split_set[4]
                    )
                    or (
                        # "they/them/theirs"
                        requested_pronouns_split[:2] == known_split_set[:2]
                        and requested_pronouns_split[2] == known_split_set[3]
                    )
                )
            ):
                matching.append(known_pronoun_set)

        if not matching:
            bot.reply(
                "I'm sorry, I don't know those pronouns. "
                "You can give me a set I don't know by formatting it "
                "subject/object/possessive-determiner/possessive-pronoun/"
                "reflexive, as in: they/them/their/theirs/themselves"
            )
            return

        requested_pronouns = matching.pop(0)
        if matching:
            disambig = " Or, if you meant one of these, please tell me: {}".format(
                ", ".join(matching)
            )

    bot.db.set_nick_value(trigger.nick, 'pronouns', requested_pronouns)
    bot.reply(
        "Thanks for telling me! I'll remember you use {}.{}".format(requested_pronouns, disambig)
    )


@plugin.command('clearpronouns')
def unset_pronouns(bot, trigger):
    """Clear pronouns for the given user."""
    bot.db.delete_nick_value(trigger.nick, 'pronouns')
    bot.reply("Okay, I'll forget your pronouns.")



"""
invite.py - Sopel Invite Plugin
Copyright © 2016, João Vanzuita, https://github.com/converge
Copyright © 2019, dgw, https://github.com/dgw
Licensed under the Eiffel Forum License 2.

https://sopel.chat
"""
from __future__ import annotations

from sopel import plugin


MIN_PRIV = plugin.HALFOP


def invite_handler(bot, sender, user, channel):
    """Common control logic for invite commands received from anywhere."""
    sender = bot.make_identifier(sender)
    user = bot.make_identifier(user)
    channel = bot.make_identifier(channel)

    # Sanity checks, in case someone reuses this function from outside the plugin
    if not sender.is_nick():
        raise ValueError("Invite sender must be a nick, not a channel.")
    if not user.is_nick():
        raise ValueError("User to invite must be a nick, not a channel.")
    if channel.is_nick():
        raise ValueError("Target channel name must not be a nick.")

    # Sopel must be in the target channel
    if channel not in bot.channels or bot.nick not in bot.channels[channel].privileges:
        bot.reply("I'm not in {}!".format(channel))
        return

    privs = bot.channels[channel].privileges

    # Sopel must have sufficient privileges in the target channel to send invites
    if privs[bot.nick] < MIN_PRIV:
        bot.reply("I don't have permission to invite anyone into {}.".format(channel))
        return

    # The sender must be in the target channel
    if sender not in privs:
        bot.reply("You're not in {}.".format(channel))
        return

    # The sender must have sufficient privileges in the target channel to send invites
    if privs[sender] < MIN_PRIV:
        bot.reply("You don't have permission to invite anyone into {}.".format(channel))
        return

    # Sopel and the sender both passed permission checks.
    # DDDDOOOO IIIITTTT
    bot.write(['INVITE', user, channel])


@plugin.command('invite')
@plugin.example('.invite jenny', user_help=True)
@plugin.example('.invite converge #sopel', user_help=True)
def invite(bot, trigger):
    """
    Invite the given user to the current channel, or (with optional
    second argument) another channel that Sopel is in.
    """
    if not trigger.group(3):
        return bot.reply("Whom should I invite?")
    user = trigger.group(3)

    if trigger.group(4):
        channel = trigger.group(4)
    else:
        if trigger.is_privmsg:
            return bot.say(
                "Channel is required ({}invite user #channel) when inviting from queries."
                .format(bot.config.core.help_prefix))
        channel = trigger.sender

    try:
        invite_handler(bot, trigger.nick, user, channel)
    except ValueError as err:
        bot.reply('%s' % err)



from __future__ import annotations

from ssl import TLSVersion

from sopel.config.types import (
    BooleanAttribute,
    ChoiceAttribute,
    FilenameAttribute,
    ListAttribute,
    NO_DEFAULT,
    SecretAttribute,
    StaticSection,
    ValidatedAttribute,
)
from sopel.tools.identifiers import Identifier


COMMAND_DEFAULT_PREFIX = r'\.'
"""Default prefix used for commands."""
COMMAND_DEFAULT_HELP_PREFIX = '.'
"""Default help prefix used in commands' usage messages."""
URL_DEFAULT_SCHEMES = ['http', 'https', 'ftp']
"""Default URL schemes allowed for URLs."""


def _parse_ssl_version(var: str) -> TLSVersion:
    """Parse an ssl_version config variable.

    :param var: The input string, e.g. "TLSv1_3".
    :return: The TLSVersion object for that version.
    """
    try:
        return TLSVersion[var]
    except KeyError:
        raise ValueError(
            "'{}' is not a valid TLSVersion. Try e.g. TLSv1_3'".format(var)
        )


def configure(config):
    """Interactively configure the bot's ``[core]`` config section.

    :param config: the bot's config object
    :type config: :class:`~.config.Config`
    """
    config.core.configure_setting('nick', 'Enter the nickname for your bot.')
    config.core.configure_setting('host', 'Enter the server to connect to.')
    config.core.configure_setting('use_ssl', 'Should the bot connect with SSL?')
    if config.core.use_ssl:
        default_port = 6697
    else:
        default_port = 6667
    config.core.configure_setting('port', 'Enter the port to connect on.',
                                  default=default_port)
    config.core.configure_setting(
        'owner', "Enter your own IRC name (or that of the bot's owner)")
    config.core.configure_setting(
        'channels',
        'Enter the channels to connect to at startup, separated by commas.'
    )
    config.core.configure_setting(
        'commands_on_connect',
        'Enter commands to perform on successful connection to server (one per \'?\' prompt).'
    )


class CoreSection(StaticSection):
    """The config section used for configuring the bot itself.

    .. important::

        All **Required** values must be specified, or Sopel will fail to start.

    .. note::

        You can use the command ``sopel configure`` to generate a config file
        with the minimal required options.

    """

    admins = ListAttribute('admins')
    """The list of people (other than the owner) who can administer the bot.

    Example:

    .. code-block:: ini

        admin =
            YourFavAdmin
            TheOtherAdmin
            YetAnotherRockstarAdmin
    """

    admin_accounts = ListAttribute('admin_accounts')
    """The list of admin accounts other than the owner's.

    Each account is allowed to administer the bot and can perform commands
    that are restricted to admins.

    Example:

    .. code-block:: ini

        admin_accounts =
            favadmin
            otheradm
            yetanotherone

    .. important::

        This should not be set for networks that do not support IRCv3 account
        capabilities. In that case, use :attr:`admins` instead.

    """

    alias_nicks = ListAttribute('alias_nicks')
    """List of alternate names users may call the bot.

    These aliases are used along with the bot's nick for ``$nick`` and
    ``$nickname`` regex substitutions.

    For example, a bot named "William" (its :attr:`nick`) could have these
    aliases:

    .. code-block:: ini

        alias_nicks =
            Bill
            Will
            Liam

    This would then allow both "William: Hi!" and "Bill: Hi!" to work with
    :func:`~sopel.plugin.nickname_command`.
    """

    antiloop_repeat_text = ValidatedAttribute(
        'antiloop_repeat_text', default='...')
    """The replacement text sent when detecting a repeated message.

    :default: ``...``

    This is equivalent to the default value:

    .. code-block:: ini

        antiloop_repeat_text = ...

    .. seealso::

        The :ref:`Loop Prevention` chapter to learn what each antiloop-related
        setting does.

    .. versionadded:: 8.0
    """

    antiloop_silent_after = ValidatedAttribute(
        'antiloop_silent_after', int, default=3)
    """How many times the anti-looping message will be sent before stopping.

    :default: ``3``

    This is equivalent to the default value:

    .. code-block:: ini

        antiloop_silent_after = 3

    .. seealso::

        The :ref:`Loop Prevention` chapter to learn what each antiloop-related
        setting does.

    .. versionadded:: 8.0
    """

    antiloop_threshold = ValidatedAttribute(
        'antiloop_threshold', int, default=5)
    """How many times a message can be repeated without anti-looping action.

    :default: ``5``

    This is equivalent to the default value:

    .. code-block:: ini

        antiloop_threshold = 5

    You can deactivate the anti-looping feature (not recommended) by setting
    this to ``0``:

    .. code-block:: ini

        antiloop_threshold = 0

    .. seealso::

        The :ref:`Loop Prevention` chapter to learn what each antiloop-related
        setting does.

    .. versionadded:: 8.0
    """

    antiloop_window = ValidatedAttribute('antiloop_window', int, default=120)
    """The time period (in seconds) checked when detecting repeated messages.

    :default: ``120``

    This is equivalent to the default value:

    .. code-block:: ini

        antiloop_window = 120

    .. seealso::

        The :ref:`Loop Prevention` chapter to learn what each antiloop-related
        setting does.

    .. versionadded:: 8.0
    """

    auth_method = ChoiceAttribute('auth_method', choices=[
        'nickserv', 'authserv', 'Q', 'sasl', 'server', 'userserv'])
    """Simple method to authenticate with the server.

    Can be one of ``nickserv``, ``authserv``, ``Q``, ``sasl``, ``server``, or
    ``userserv``.

    This allows only a single authentication method; to use both a server-based
    authentication method *and* a nick-based authentication method, see
    :attr:`server_auth_method` and :attr:`nick_auth_method`.

    For more information about these methods, see :ref:`Authentication`.

    .. note::
        If this is specified, :attr:`nick_auth_method` will be ignored, and this
        value will override :attr:`server_auth_method`.
    """

    auth_password = SecretAttribute('auth_password')
    """The password to use to authenticate with the :attr:`auth_method`.

    See :ref:`Authentication`.
    """

    auth_target = ValidatedAttribute('auth_target')
    """Target for authentication.

    :default:
        * ``NickServ`` if using the ``nickserv`` :attr:`auth_method`
        * ``UserServ`` if using the ``userserv`` :attr:`auth_method`
        * ``PLAIN`` if using the ``sasl`` :attr:`auth_method`

    The nickname of the NickServ or UserServ service, or the name of the
    desired SASL mechanism, if :attr:`auth_method` is set to one of these
    methods. For SASL, the ``EXTERNAL`` option is available in case the IRC
    network requires it (e.g. for CertFP using :attr:`client_cert_file`). This
    value is otherwise ignored.

    See :ref:`Authentication`.
    """

    auth_username = ValidatedAttribute('auth_username')
    """The user/account name to use when authenticating.

    Required for an :attr:`auth_method` of ``authserv``, ``Q``, or
    ``userserv`` — otherwise ignored.

    See :ref:`Authentication`.
    """

    auto_url_schemes = ListAttribute(
        'auto_url_schemes',
        strip=True,
        default=URL_DEFAULT_SCHEMES)
    """List of URL schemes that will trigger URL callbacks.

    :default: ``['http', 'https', 'ftp']``

    Used by the URL callbacks feature to call plugins when links are posted in
    chat; see the :func:`sopel.plugin.url` decorator.

    The default value allows ``http``, ``https``, and ``ftp``. It is equivalent
    to this configuration example:

    .. code-block:: ini

        auto_url_schemes =
            http
            https
            ftp

    """

    bind_host = ValidatedAttribute('bind_host')
    """Bind the connection to a specific IP.

    :default: ``0.0.0.0`` (all interfaces)

    This is equivalent to the default value:

    .. code-block:: ini

        bind_host = 0.0.0.0

    """

    ca_certs = FilenameAttribute('ca_certs')
    """The path to the CA certs ``PEM`` file.

    Example:

    .. code-block:: ini

        ca_certs = /etc/ssl/certs/ca-certificates.crt

    If not specified, the system default will be used.

    If the given value is not an absolute path, it will be interpreted relative
    to the directory containing the config file with which Sopel was started.
    """

    channels = ListAttribute('channels')
    """List of channels for the bot to join when it connects.

    If a channel key needs to be provided, separate it from the channel name
    with a space:

    .. code-block:: ini

        channels =
            "#channel"
            "#logs"
            &rare_prefix_channel
            "#private password"

    .. important::

        If you edit the config file manually, make sure to wrap each line
        starting with a ``#`` in double quotes, as shown in the example above.
        An unquoted ``#`` denotes a comment, which will be ignored by Sopel's
        configuration parser.

    """

    client_cert_file = FilenameAttribute('client_cert_file')
    """Filesystem path to a certificate file for CertFP.

    This is expected to be a ``.pem`` file containing both the certificate and
    private key. Most networks that support CertFP will give instructions for
    generating this, typically using OpenSSL.

    Some networks may refer to this authentication method as SASL EXTERNAL.

    .. versionadded:: 8.0
    """

    commands_on_connect = ListAttribute('commands_on_connect')
    """A list of commands to send upon successful connection to the IRC server.

    Each line is a message that will be sent to the server once connected,
    in the order they are defined:

    .. code-block:: ini

        commands_on_connect =
            PRIVMSG Q@CServe.quakenet.org :AUTH my_username MyPassword,@#$%!
            PRIVMSG MyOwner :I'm here!

    ``$nickname`` can be used in a command as a placeholder, and will be
    replaced with the bot's :attr:`nick`. For example, if the bot's nick is
    ``Sopel``, ``MODE $nickname +Xxw`` will be expanded to ``MODE Sopel +Xxw``.

    .. versionadded:: 7.0
    """

    db_driver = ValidatedAttribute('db_driver')
    """The driver to use for connecting to the database.

    This is optional, but can be specified if user wants to use a different
    driver than the default for the chosen :attr:`db_type`.

    .. seealso::

        Refer to :ref:`SQLAlchemy's documentation <engines_toplevel>` for more
        information.

    """

    db_filename = ValidatedAttribute('db_filename')
    """The filename for Sopel's database.

    Used only for SQLite. Ignored for all other :attr:`db_type` values.
    """

    db_host = ValidatedAttribute('db_host')
    """The host for Sopel's database.

    Ignored when using SQLite.
    """

    db_name = ValidatedAttribute('db_name')
    """The name of Sopel's database.

    Ignored when using SQLite.
    """

    db_pass = SecretAttribute('db_pass')
    """The password for Sopel's database.

    Ignored when using SQLite.
    """

    db_port = ValidatedAttribute('db_port')
    """The port for Sopel's database.

    Ignored when using SQLite.
    """

    db_type = ChoiceAttribute('db_type', choices=[
        'sqlite', 'mysql', 'postgres', 'mssql', 'oracle', 'firebird', 'sybase'], default='sqlite')
    """The type of database Sopel should connect to.

    :default: ``sqlite`` (part of Python's standard library)

    The full list of values Sopel recognizes is:

    * ``firebird``
    * ``mssql``
    * ``mysql``
    * ``oracle``
    * ``postgres``
    * ``sqlite``
    * ``sybase``

    Here are the additional PyPI packages you may need to install to use one of
    the most commonly requested alternatives:

    mysql
      ``pip install mysql-python`` (Python 2)

      ``pip install mysqlclient`` (Python 3)

    postgres
      ``pip install psycopg2``

    mssql
      ``pip install pymssql``

    This is equivalent to the default value:

    .. code-block:: ini

        db_type = sqlite

    .. seealso::

        Refer to :ref:`SQLAlchemy's documentation <dialect_toplevel>` for more
        information about the different dialects it supports.

    .. note::

        Plugins originally written for Sopel 6.x and older *might* not work
        correctly with ``db_type``\\s other than ``sqlite``.

    """

    db_url = ValidatedAttribute('db_url')
    """A raw database URL.

    If this option is present, Sopel will ignore **all** other ``db_*``
    settings and use this option's value only.

    .. note::

        Specifying this option via the ``SOPEL_CORE_DB_URL`` :ref:`environment
        variable <Overriding individual settings>` may prove especially useful
        in certain cloud environments, avoiding the need to split a database
        URI provided by the platform at runtime into its components with a
        startup script.

    .. versionadded:: 8.0
    """

    db_user = ValidatedAttribute('db_user')
    """The user for Sopel's database.

    Ignored when using SQLite.
    """

    default_time_format = ValidatedAttribute('default_time_format',
                                             default='%Y-%m-%d - %T %Z')
    """The default format to use for time in messages.

    :default: ``%Y-%m-%d - %T %Z``

    Used when plugins format times with :func:`sopel.tools.time.format_time`.

    This is equivalent to the default value:

    .. code-block:: ini

        default_time_format = %Y-%m-%d - %T %Z

    .. seealso::

        Time format reference is available in the documentation for Python's
        :func:`time.strftime` function.

    """

    default_timezone = ValidatedAttribute('default_timezone', default='UTC')
    """The default timezone to use for time in messages.

    :default: ``UTC``

    .. highlight:: ini

    Used when plugins format times with :func:`sopel.tools.time.format_time`.

    For example, to make Sopel fall back on British time::

        default_timezone = Europe/London

    And this is equivalent to the default value::

        default_timezone = UTC

    """

    enable = ListAttribute('enable')
    """A list of the only plugins you want to enable.

    .. highlight:: ini

    If set, Sopel will *only* load the plugins named here. All other available
    plugins will be ignored::

        enable =
            url
            xkcd
            help

    In that case, only the ``url``, ``xkcd``, and ``help`` plugins will be
    enabled and loaded by Sopel.

    To load *all* available plugins, clear this setting by removing it, or
    by making it empty::

        enable =

    To disable only a few plugins, see :attr:`exclude`.

    .. seealso::

        The :ref:`Plugins` chapter for an overview of all plugin-related
        settings.

    """

    exclude = ListAttribute('exclude')
    """A list of plugins which should not be loaded.

    .. highlight:: ini

    If set, Sopel will load all available plugins *except* those named here::

        exclude =
            url
            calc
            meetbot

    In that case, ``url``, ``calc``, and ``meetbot`` will be excluded, and they
    won't be loaded by Sopel.

    A plugin named both here and in :attr:`enable` **will not** be loaded;
    :attr:`exclude` takes priority.

    .. seealso::

        The :ref:`Plugins` chapter for an overview of all plugin-related
        settings.

    """

    extra = ListAttribute('extra')
    """A list of other directories in which to search for plugin files.

    Example:

    .. code-block:: ini

        extra =
            /home/myuser/custom-sopel-plugins/
            /usr/local/lib/ad-hoc-plugins/

    .. seealso::

        The :ref:`Plugins` chapter for an overview of all plugin-related
        settings.

    """

    flood_burst_lines = ValidatedAttribute('flood_burst_lines', int, default=4)
    """How many messages can be sent in burst mode.

    :default: ``4``

    This is equivalent to the default value:

    .. code-block:: ini

        flood_burst_lines = 4

    .. seealso::

        The :ref:`Flood Prevention` chapter to learn what each flood-related
        setting does.

    .. versionadded:: 7.0
    """

    flood_empty_wait = ValidatedAttribute('flood_empty_wait', float, default=0.7)
    """How long to wait between sending messages when not in burst mode, in seconds.

    :default: ``0.7``

    This is equivalent to the default value:

    .. code-block:: ini

        flood_empty_wait = 0.7

    .. seealso::

        The :ref:`Flood Prevention` chapter to learn what each flood-related
        setting does.

    .. versionadded:: 7.0
    """

    flood_max_wait = ValidatedAttribute('flood_max_wait', float, default=2)
    """How many seconds to wait at most when flood protection kicks in.

    :default: ``2``

    .. note::

        If the maximum wait is 0, flood protection is effectively disabled.

    This is equivalent to the default value:

    .. code-block:: ini

        flood_max_wait = 2

    .. seealso::

        The :ref:`Flood Prevention` chapter to learn what each flood-related
        setting does.

    .. versionadded:: 7.1
    """

    flood_penalty_ratio = ValidatedAttribute('flood_penalty_ratio',
                                             float,
                                             default=1.4)
    """Ratio of the message length used to compute the added wait penalty.

    :default: ``1.4``

    Messages longer than :attr:`flood_text_length` will get an added
    wait penalty (in seconds) that will be computed like this::

        overflow = max(0, (len(text) - flood_text_length))
        rate = flood_text_length * flood_penalty_ratio
        penalty = overflow / rate

    .. note::

        If the penalty ratio is 0, this penalty will be disabled.

    This is equivalent to the default value:

    .. code-block:: ini

        flood_penalty_ratio = 1.4

    .. seealso::

        The :ref:`Flood Prevention` chapter to learn what each flood-related
        setting does.

    .. versionadded:: 7.1
    """

    flood_refill_rate = ValidatedAttribute('flood_refill_rate', int, default=1)
    """How quickly burst mode recovers, in messages per second.

    :default: ``1``

    This is equivalent to the default value:

    .. code-block:: ini

        flood_refill_rate = 1

    .. seealso::

        The :ref:`Flood Prevention` chapter to learn what each flood-related
        setting does.

    .. versionadded:: 7.0
    """

    flood_text_length = ValidatedAttribute('flood_text_length', int, default=50)
    """Length of text at which an extra wait penalty is added.

    :default: ``50``

    Messages longer than this (in bytes) get an added wait penalty if the
    flood protection limit is reached.

    This is equivalent to the default value:

    .. code-block:: ini

        flood_text_length = 50

    .. seealso::

        The :ref:`Flood Prevention` chapter to learn what each flood-related
        setting does.

    .. versionadded:: 7.1
    """

    help_prefix = ValidatedAttribute('help_prefix',
                                     default=COMMAND_DEFAULT_HELP_PREFIX)
    """The prefix to use in help output.

    :default: ``.``

    This is equivalent to the default value:

    .. code-block:: ini

        help_prefix = .

    If :attr:`prefix` is changed from the default, this setting **must** be
    updated to reflect the prefix your bot will actually respond to, or the
    built-in ``help`` functionality will provide incorrect example usage.
    """

    @property
    def homedir(self):
        """The directory in which various files are stored at runtime.

        Specifying the ``homedir`` option is useful for e.g. :doc:`running Sopel
        as a system service </run/service>`.

        If not set, the config file's parent directory will be used. This value
        cannot be changed at runtime.
        """
        return self._parent.homedir

    host = ValidatedAttribute('host', default='irc.libera.chat')
    """The IRC server to connect to.

    :default: ``irc.libera.chat``

    **Required**:

    .. code-block:: ini

        host = irc.libera.chat

    """

    host_blocks = ListAttribute('host_blocks')
    """A list of hostnames which Sopel should ignore.

    Sopel will :ref:`ignore <Ignoring Users>` messages from any user whose
    connection hostname matches one of these values.
    :ref:`Regular expression syntax <re-syntax>` is supported, so remember to
    escape special characters:

    .. code-block:: ini

        host_blocks =
            (.+\\.)*domain\\.com

    .. seealso::

        The :attr:`nick_blocks` list can be used to block users by their nick.

    .. note::

        :ref:`Plugin callables` with the :func:`.plugin.unblockable` decorator
        run regardless of matching ``*_blocks`` entries.

        We are working toward a better block system; see :issue:`1355` for more
        information and updates.

    """

    log_raw = BooleanAttribute('log_raw', default=False)
    """Whether a log of raw lines as sent and received should be kept.

    :default: ``no``

    To enable this logging:

    .. code-block:: ini

        log_raw = yes

    .. seealso::

        The :ref:`Raw Logs` chapter.

    """

    logdir = FilenameAttribute('logdir', directory=True, default='logs')
    """Directory in which to place logs.

    :default: ``logs``

    If the given value is not an absolute path, it will be interpreted relative
    to the :attr:`homedir` of the config file with which Sopel was started.

    .. seealso::

        The :ref:`Logging` chapter.

    """

    logging_channel = ValidatedAttribute('logging_channel', Identifier)
    """The channel to send logging messages to.

    .. seealso::

        The :ref:`Log to a Channel` chapter.

    """

    logging_channel_datefmt = ValidatedAttribute('logging_channel_datefmt')
    """The format string to use for timestamps in IRC channel logs.

    If not specified, this falls back to using :attr:`logging_datefmt`.

    .. seealso::

        Time format reference is available in the documentation for Python's
        :func:`time.strftime` function.

        For more information about logging, see :ref:`Log to a Channel`.

    .. versionadded:: 7.0
    """

    logging_channel_format = ValidatedAttribute('logging_channel_format')
    """The logging format string to use in IRC channel logs.

    If not specified, this falls back to using :attr:`logging_format`.

    .. seealso::

        The :ref:`Log to a Channel` chapter.

    .. versionadded:: 7.0
    """

    logging_channel_level = ChoiceAttribute('logging_channel_level',
                                            ['CRITICAL', 'ERROR', 'WARNING',
                                             'INFO'],
                                            'WARNING')
    """The lowest severity of logs to display in IRC channel logs.

    .. seealso::

        The :ref:`Log to a Channel` chapter.

    .. versionadded:: 7.0
    .. versionchanged:: 8.0

        No longer uses the value of :attr:`logging_level` if not set. Removed
        ``DEBUG`` from the available choices; setting it caused the bot to get
        caught in an ever-increasing flood prevention loop.

    """

    logging_datefmt = ValidatedAttribute('logging_datefmt')
    """The format string to use for timestamps in logs.

    If not set, the ``datefmt`` argument is not provided, and :mod:`logging`
    will use the Python default.

    .. seealso::

        Time format reference is available in the documentation for Python's
        :func:`time.strftime` function.

    .. versionadded:: 7.0
    """

    logging_format = ValidatedAttribute(
        'logging_format',
        default='[%(asctime)s] %(name)-20s %(levelname)-8s - %(message)s')
    """The logging format string to use for logs.

    :default: ``[%(asctime)s] %(name)-20s %(levelname)-8s - %(message)s``

    The default log line format will output the timestamp, the package that
    generated the log line, the log level of the line, and (finally) the actual
    message. For example::

        [2019-10-21 12:47:44,272] sopel.irc            INFO     - Connected.

    This is equivalent to the default value:

    .. code-block:: ini

       logging_format = [%(asctime)s] %(name)-20s %(levelname)-8s - %(message)s

    .. seealso::

        Python's logging format documentation: :ref:`logrecord-attributes`

    .. versionadded:: 7.0
    """

    logging_level = ChoiceAttribute('logging_level',
                                    ['CRITICAL', 'ERROR', 'WARNING', 'INFO',
                                     'DEBUG'],
                                    'INFO')
    """The lowest severity of logs to display.

    :default: ``INFO``

    Valid values sorted by increasing verbosity:

    * ``CRITICAL``
    * ``ERROR``
    * ``WARNING``
    * ``INFO``
    * ``DEBUG``

    For example to log only at WARNING level and above:

    .. code-block:: ini

        logging_level = WARNING

    """

    modes = ValidatedAttribute('modes')
    """User modes to be set on connection.

    Include only the mode letters; this value is automatically prefixed with
    ``+`` before Sopel sends the MODE command to IRC.

    .. versionchanged:: 8.0.0

        Now empty by default. Previous default was ``B``, which has been
        dropped in favor of the formal `bot mode specification`__.

        .. __: https://ircv3.net/specs/extensions/bot-mode

    """

    name = ValidatedAttribute('name', default='Sopel: https://sopel.chat/')
    """The "real name" of your bot for ``WHOIS`` responses.

    :default: ``Sopel: https://sopel.chat/``
    """

    nick = ValidatedAttribute('nick', default='Sopel')
    """The nickname for the bot.

    :default: ``Sopel``

    **Required**:

    .. code-block:: ini

        nick = Sopel

    """

    nick_auth_method = ChoiceAttribute('nick_auth_method', choices=[
        'nickserv', 'authserv', 'Q', 'userserv'])
    """The nick authentication method.

    Can be one of ``nickserv``, ``authserv``, ``Q``, or ``userserv``.

    .. seealso::

        The :ref:`Authentication` chapter for more details.

    .. versionadded:: 7.0
    """

    nick_auth_password = SecretAttribute('nick_auth_password')
    """The password to use to authenticate the bot's nick.

    .. seealso::

        The :ref:`Authentication` chapter for more details.

    .. versionadded:: 7.0
    """

    nick_auth_target = ValidatedAttribute('nick_auth_target')
    """The target user for nick authentication.

    :default: ``NickServ`` for ``nickserv`` authentication; ``UserServ`` for
              ``userserv`` authentication

    May not apply, depending on the chosen :attr:`nick_auth_method`.

    .. seealso::

        The :ref:`Authentication` chapter for more details.

    .. versionadded:: 7.0
    """

    nick_auth_username = ValidatedAttribute('nick_auth_username')
    """The username/account to use for nick authentication.

    :default: the value of :attr:`nick`

    May not apply, depending on the chosen :attr:`nick_auth_method`.

    .. seealso::

        The :ref:`Authentication` chapter for more details.

    .. versionadded:: 7.0
    """

    nick_blocks = ListAttribute('nick_blocks')
    """A list of nicks which Sopel should ignore.

    Sopel will :ref:`ignore <Ignoring Users>` messages from any user whose
    nickname matches one of these values.
    :ref:`Regular expression syntax <re-syntax>` is supported, so remember to
    escape special characters:

    .. code-block:: ini

        nick_blocks =
            ExactNick
            _*RegexMatch_*

    .. seealso::

        The :attr:`host_blocks` list can be used to block users by their host.

    .. note::

        :ref:`Plugin callables` with the :func:`.plugin.unblockable` decorator
        run regardless of matching ``*_blocks`` entries.

        We are working toward a better block system; see :issue:`1355` for more
        information and updates.

    """

    not_configured = BooleanAttribute('not_configured', default=False)
    """For package maintainers. Not used in normal configurations.

    :default: ``False``

    This allows software packages to install a default config file, with this
    option set to ``True``, so that commands to start, stop, or restart the bot
    won't work until the bot has been properly configured.
    """

    owner = ValidatedAttribute('owner', default=NO_DEFAULT)
    """The IRC name of the owner of the bot.

    **Required** even if :attr:`owner_account` is set.
    """

    owner_account = ValidatedAttribute('owner_account')
    """The services account name of the owner of the bot.

    This should only be set on networks which support IRCv3 account
    capabilities.
    """

    pid_dir = FilenameAttribute('pid_dir', directory=True, default='.')
    """The directory in which to put the file Sopel uses to track its process ID.

    :default: ``.``

    If the given value is not an absolute path, it will be interpreted relative
    to the directory containing the config file with which Sopel was started.

    You probably do not need to change this unless you're managing Sopel with
    ``systemd`` or similar.
    """

    port = ValidatedAttribute('port', int, default=6697)
    """The port to connect on.

    :default: ``6697``

    .. highlight:: ini

    **Required**::

        port = 6697

    Or if SSL is disabled::

        port = 6667
        use_ssl = false

    """

    prefix = ValidatedAttribute('prefix', default=COMMAND_DEFAULT_PREFIX)
    """The prefix to add to the beginning of commands as a regular expression.

    :default: ``\\.``

    .. highlight:: ini

    **Required**::

        prefix = \\.

    With the default value, users will invoke commands like this:

    .. code-block:: irc

        <nick> .help

    Since it's a regular expression, you can use multiple prefixes::

        prefix = \\.|\\?

    .. important::

        As the prefix is a regular expression, don't forget to escape it when
        necessary. It is not recommended to use capturing groups, as it
        **will** create problems with argument parsing for commands.

    .. note::

        Remember to change the :attr:`help_prefix` value accordingly::

            prefix = \\?
            help_prefix = ?

        In that example, users will invoke commands like this:

        .. code-block:: irc

            <nick> ?help xkcd
            <Sopel> ?xkcd - Finds an xkcd comic strip
            <Sopel> Takes one of 3 inputs:
            [...]

    """

    reply_errors = BooleanAttribute('reply_errors', default=True)
    """Whether to reply to the sender of a message that triggered an error.

    :default: ``True``

    If ``True``, Sopel will send information about the triggered exception to
    the sender of the message that caused the error.

    If ``False``, Sopel will only log the error and will appear to fail
    silently from the triggering IRC user's perspective.
    """

    server_auth_method = ChoiceAttribute('server_auth_method',
                                         choices=['sasl', 'server'])
    """The server authentication method.

    Can be ``sasl`` or ``server``.

    .. versionadded:: 7.0
    """

    server_auth_password = SecretAttribute('server_auth_password')
    """The password to use to authenticate with the server.

    .. versionadded:: 7.0
    """

    server_auth_sasl_mech = ValidatedAttribute('server_auth_sasl_mech')
    """The SASL mechanism.

    :default: ``PLAIN``

    ``EXTERNAL`` is also supported, e.g. for using :attr:`client_cert_file` to
    authenticate via CertFP.

    .. versionadded:: 7.0
    .. versionchanged:: 8.0
        Added support for SASL EXTERNAL mechanism.
    """

    server_auth_username = ValidatedAttribute('server_auth_username')
    """The username/account to use to authenticate with the server.

    .. versionadded:: 7.0
    """

    throttle_join = ValidatedAttribute('throttle_join', int, default=0)
    """Slow down the initial join of channels to prevent getting kicked.

    :default: ``0``

    Sopel will only join this many channels at a time, sleeping for a second
    between each batch to avoid getting kicked for joining too quickly. This is
    unnecessary on most networks.

    If not set, or set to 0, Sopel won't slow down the initial join.

    In this example, Sopel will try to join 4 channels at a time:

    .. code-block:: ini

        throttle_join = 4

    .. seealso::

        :attr:`throttle_wait` controls Sopel's waiting time between joining
        batches of channels.

    """

    throttle_wait = ValidatedAttribute('throttle_wait', int, default=1)
    """Time in seconds Sopel waits between joining batches of channels.

    :default: ``1``

    In this example:

    .. code-block:: ini

        throttle_wait = 5
        throttle_join = 2

    Sopel will join 2 channels every 5s.

    If :attr:`throttle_join` is ``0``, this setting has no effect.

    .. seealso::

        :attr:`throttle_join` controls channel batch size.

    """

    timeout = ValidatedAttribute('timeout', int, default=120)
    """The number of seconds acceptable since the last message before timing out.

    :default: ``120``

    You can change the timeout like this:

    .. code-block:: ini

        # increase to 200 seconds
        timeout = 200

    """

    timeout_ping_interval = ValidatedAttribute('timeout_ping_interval',
                                               int,
                                               default=0)
    """The number of seconds before sending a PING command to the server.

    :default: (auto)

    The default value is made to send at least 2 PINGs before the bot thinks
    there is a timeout, given that :attr:`timeout` is 120s by default:

    * t+54s: first PING
    * t+108s: second PING
    * t+120s: if no response, then a timeout is detected

    This makes the timeout detection more lenient and forgiving, by allowing a
    12s window for the server to send something that will prevent a timeout.

    You can change the PING interval like this:

    .. code-block:: ini

        # timeout up to 250s
        timeout = 250
        # PING every 80s (about 3 times every 240s + 10s window)
        timeout_ping_interval = 80

    .. note::

        Internally, the default value is 0 and the value used will be
        automatically calculated as 45% of the :attr:`timeout`::

            ping_interval = timeout * 0.45

        So for a timeout of 120s it's a PING every 54s. For a timeout of 250s
        it's a PING every 112.5s.

    """

    ssl_ciphers = ListAttribute(
        'ssl_ciphers',
        default=[
            "ECDHE-ECDSA-AES128-GCM-SHA256",
            "ECDHE-RSA-AES128-GCM-SHA256",
            "ECDHE-ECDSA-AES256-GCM-SHA384",
            "ECDHE-RSA-AES256-GCM-SHA384",
            "ECDHE-ECDSA-CHACHA20-POLY1305",
            "ECDHE-RSA-CHACHA20-POLY1305",
            "DHE-RSA-AES128-GCM-SHA256",
            "DHE-RSA-AES256-GCM-SHA384",
        ],
    )
    """The cipher suites enabled for SSL/TLS connections.

    :default: ``ECDHE-ECDSA-AES128-GCM-SHA256
                ECDHE-RSA-AES128-GCM-SHA256
                ECDHE-ECDSA-AES256-GCM-SHA384
                ECDHE-RSA-AES256-GCM-SHA384
                ECDHE-ECDSA-CHACHA20-POLY1305
                ECDHE-RSA-CHACHA20-POLY1305
                DHE-RSA-AES128-GCM-SHA256
                DHE-RSA-AES256-GCM-SHA384``

    The default is the `Mozilla Intermediate configuration`__ as of February
    2022. This parameter is in OpenSSL format; see the `ciphers manual page`__
    for your version. This setting cannot be used to configure TLS 1.3.

    .. __: https://wiki.mozilla.org/Security/Server_Side_TLS
    .. __: https://www.openssl.org/docs/manmaster/man1/ciphers.html

    .. code-block:: ini

        ssl_ciphers =
            ECDSA+AES256
            !SHA1

    .. note::

        Cipher selection is also subject to the available SSL/TLS versions;
        see :attr:`ssl_minimum_version`.

    """

    ssl_minimum_version = ValidatedAttribute(
        'ssl_minimum_version',
        default=TLSVersion.TLSv1_2,
        parse=_parse_ssl_version,
        serialize=lambda ver: ver.name,
    )
    """The minimum version allowed for SSL/TLS connections.

    :default: ``TLSv1_2``

    You should not set this lower than the default unless the server does not
    support modern versions. Set this to a :class:`~ssl.TLSVersion` value,
    e.g. ``TLSv1`` or ``TLSv1_3``.

    .. code-block:: ini

        ssl_minimum_version = TLSv1_3

    .. note::

        To use insecure SSL/TLS versions, you may also need to enable
        insecure cipher suites. See :attr:`ssl_ciphers`.

    """

    use_ssl = BooleanAttribute('use_ssl', default=True)
    """Whether to use a SSL/TLS encrypted connection.

    :default: ``True``

    Example with SSL off:

    .. code-block:: ini

        use_ssl = false

    """

    user = ValidatedAttribute('user', default='sopel')
    """The "user" for your bot (the part before the ``@`` in the hostname).

    :default: ``sopel``

    **Required**:

    .. code-block:: ini

        user = sopel

    """

    verify_ssl = BooleanAttribute('verify_ssl', default=True)
    """Whether to require a trusted certificate for encrypted connections.

    :default: ``True``

    Example with SSL on:

    .. code-block:: ini

        use_ssl = true
        verify_ssl = true

    """



"""Deprecation module for Sopel developers and plugin authors.

.. versionadded:: 8.0

    Previously in :mod:`sopel.tools`, the :func:`deprecated` function has been
    moved to this newly created module, as it can be used in every part of the
    Sopel codebase, including :mod:`sopel.tools` itself.

"""
from __future__ import annotations

import functools
import inspect
import logging
import traceback
from typing import Callable

from packaging.version import parse as parse_version

from sopel import __version__


def deprecated(
    reason: str | Callable | None = None,
    version: str | None = None,
    removed_in: str | None = None,
    warning_in: str | None = None,
    stack_frame: int = -1,
    func: Callable | None = None,
) -> Callable:
    """Decorator to mark deprecated functions in Sopel's API

    :param reason: optional text added to the deprecation warning
    :param version: optional version number when the decorated function
                    is deprecated
    :param removed_in: optional version number when the deprecated function
                       will be removed
    :param warning_in: optional version number when the decorated function
                       should start emitting a warning when called
    :param stack_frame: optional stack frame to output; defaults to
                        ``-1``; should almost always be negative
    :param func: deprecated function
    :return: a callable that depends on how the decorator is called; either
             the decorated function, or a decorator with the appropriate
             parameters

    Any time the decorated ``func`` is called, a deprecation warning will be
    logged, with the last frame of the traceback. The optional ``warning_in``
    argument suppresses the warning on Sopel versions older than that, allowing
    for multi-stage deprecation timelines.

    The decorator can be used with or without arguments::

        from sopel.lifecycle import deprecated

        @deprecated
        def func1():
            print('func 1')

        @deprecated()
        def func2():
            print('func 2')

        @deprecated(reason='obsolete', version='7.0', removed_in='8.0')
        def func3():
            print('func 3')

    which will output the following in a console::

        >>> func1()
        Deprecated: func1
        File "<stdin>", line 1, in <module>
        func 1
        >>> func2()
        Deprecated: func2
        File "<stdin>", line 1, in <module>
        func 2
        >>> func3()
        Deprecated since 7.0, will be removed in 8.0: obsolete
        File "<stdin>", line 1, in <module>
        func 3

    The ``stack_frame`` argument can be used to choose which stack frame is
    logged along with the message text. By default, this decorator logs the
    most recent stack frame (the last entry in the list, ``-1``), corresponding
    to where the decorated function itself was called. However, in certain
    cases such as deprecating conditional behavior within an object
    constructor, it can be useful to show a less recent stack frame instead.

    .. note::

        This decorator can be also used on callables that are not functions,
        such as classes and callable objects.

    .. versionadded:: 7.0
        Parameters ``reason``, ``version``, and ``removed_in``.

    .. versionadded:: 7.1
        The ``warning_in`` and ``stack_frame`` parameters.

    .. versionchanged:: 8.0
        Moved out of :mod:`sopel.tools` to resolve circular dependency issues.

    """
    if not any([reason, version, removed_in, warning_in, func]):
        # common usage: @deprecated()
        return deprecated

    if callable(reason):
        # common usage: @deprecated
        return deprecated(func=reason)

    if func is None:
        # common usage: @deprecated(message, version, removed_in)
        def decorator(func):
            return deprecated(
                reason, version, removed_in, warning_in, stack_frame, func)
        return decorator

    # now, we have everything we need to have:
    # - message is not a callable (could be None)
    # - func is not None
    # - version and removed_in can be None but that's OK
    # so now we can return the actual decorated function

    message = reason or getattr(func, '__name__', '<anonymous-function>')

    template = 'Deprecated: {message}'
    if version and removed_in:
        template = (
            'Deprecated since {version}, '
            'will be removed in {removed_in}: '
            '{message}')
    elif version:
        template = 'Deprecated since {version}: {message}'
    elif removed_in:
        template = 'Deprecated, will be removed in {removed_in}: {message}'

    text = template.format(
        message=message, version=version, removed_in=removed_in)

    @functools.wraps(func)
    def deprecated_func(*args, **kwargs):
        warn_ver = warning_in and parse_version(warning_in).release
        this_ver = parse_version(__version__).release

        if not (warn_ver and warn_ver >= this_ver):
            original_frame = inspect.stack()[-stack_frame]
            mod = inspect.getmodule(original_frame[0])
            module_name = None
            if mod:
                module_name = mod.__name__
            if module_name:
                if module_name.startswith('sopel.'):
                    # core, or core plugin
                    logger = logging.getLogger(module_name)
                else:
                    # probably a plugin; try to handle most cases sanely
                    if module_name.startswith('sopel_modules.'):
                        # namespace package plugins have a prefix, obviously
                        # they will use Sopel's namespace; other won't
                        module_name = module_name.replace(
                            'sopel_modules.',
                            'sopel.externals.',
                            1,
                        )
                    logger = logging.getLogger(module_name)
            else:
                # don't know the module/plugin name, but we want to make sure
                # the log line is still output, so just get *something*
                logger = logging.getLogger(__name__)

            # Format only the desired stack frame
            trace = traceback.extract_stack()
            trace_frame = traceback.format_list(trace[:-1])[stack_frame][:-1]

            # Warn the user
            logger.warning(text + "\n" + trace_frame)

        return func(*args, **kwargs)

    return deprecated_func
